from numbers import Integral
from typing import Union, Optional, List, Iterable

import numpy as np

from deeptime.base import Model
from ._output_model import OutputModel, DiscreteOutputModel
from deeptime.markov import sample
from ._hmm_bindings.util import viterbi as viterbi_impl, forward as forward_impl
from ...util.types import ensure_dtraj_list, ensure_array


class HiddenMarkovModel(Model):
    """ Hidden Markov state model consisting of a transition model
    (:class:`MSM <deeptime.markov.msm.MarkovStateModel>`) on the hidden states, an
    :class:`output model <OutputModel>` which maps from the hidden states to a distribution of observable states,
    and optionally an initial distribution on the hidden states. Some properties require a crisp assignment to
    states in the observable space, in which case only a discrete output model can be used.

    Parameters
    ----------
    transition_model : (m,m) ndarray or MarkovStateModel
        Transition matrix for hidden (macro) states
    output_model : (m,n) ndarray or OutputModel
        observation probability matrix from hidden to observable (micro) states or OutputModel instance which yields
        the mapping from hidden to observable state.
    initial_distribution : (m,) ndarray, optional, default=None
        Initial distribution of the hidden (macro) states. Default is uniform.
    likelihoods : (k,) ndarray, optional, default=None
        Likelihood progression of the HMM as it was trained for k iterations with Baum-Welch.
    state_probabilities : list of ndarray, optional, default=None
        List of state probabilities for each trajectory that the model was trained on (gammas).
    initial_count : ndarray, optional, default=None
        Initial counts of the hidden (macro) states, computed from the gamma output of the Baum-Welch algorithm
    hidden_state_trajectories : list of ndarray, optional, default=None
        When estimating the HMM the data's most likely hidden state trajectory is determined and can be saved
        with the model by providing this argument.
    stride : int or str('effective'), optional, default=1
        Stride which was used to subsample discrete trajectories while estimating a HMM. Can either be an integer
        value which determines the offset or 'effective', which makes an estimate of a stride at which subsequent
        discrete trajectory elements are uncorrelated.
    observation_symbols : array_like, optional, default=None
        Sorted unique symbols in observations. If None, it is assumed that all possible observations are made
        and the state symbols are set to an iota range over the number of observation states.
    observation_symbols_full : array_like, optional, default=None
        Full set of symbols in observations. If None, it is assumed to coincide with observation_symbols.

    References
    ----------
    .. bibliography:: /references.bib
        :style: unsrt
        :filter: docname in docnames
        :keyprefix: hmm-

    See Also
    --------
    init.discrete.metastable_from_data : initial guess from data with discrete output model
    init.discrete.metastable_from_msm : initial guess from MSM with discrete output model
    init.gaussian.from_data : initial guess from data with Gaussian output model
    MaximumLikelihoodHMM : maximum likelihood estimation of HMMs
    BayesianHMM : Bayesian sampling of models for confidences.
    """

    def __init__(self, transition_model, output_model: Union[np.ndarray, OutputModel],
                 initial_distribution: Optional[np.ndarray] = None, likelihoods: Optional[np.ndarray] = None,
                 state_probabilities: Optional[List[np.ndarray]] = None, initial_count: Optional[np.ndarray] = None,
                 hidden_state_trajectories: Optional[Iterable[np.ndarray]] = None, stride: Union[int, str] = 1,
                 observation_symbols: Optional[np.ndarray] = None,
                 observation_symbols_full: Optional[np.ndarray] = None):
        super().__init__()
        if isinstance(transition_model, np.ndarray):
            from deeptime.markov.msm import MarkovStateModel
            transition_model = MarkovStateModel(transition_model)
        if isinstance(output_model, np.ndarray):
            output_model = DiscreteOutputModel(output_model)
        if transition_model.n_states != output_model.n_hidden_states:
            raise ValueError("Transition model must describe hidden states")
        if initial_distribution is None:
            # uniform
            initial_distribution = np.ones(transition_model.n_states) / transition_model.n_states
        if initial_distribution.shape[0] != transition_model.n_states:
            raise ValueError("Initial distribution over hidden states must be of length {}"
                             .format(transition_model.n_states))
        self._transition_model = transition_model
        self._output_model = output_model
        self._initial_distribution = initial_distribution
        self._likelihoods = likelihoods
        self._state_probabilities = state_probabilities
        self._initial_count = initial_count
        self._hidden_state_trajectories = hidden_state_trajectories
        if observation_symbols is None and output_model.n_observable_states >= 0:
            observation_symbols = np.arange(output_model.n_observable_states)
            observation_symbols_full = observation_symbols
        self._observation_symbols = observation_symbols
        self._observation_symbols_full = observation_symbols_full
        if not (isinstance(stride, Integral) or (isinstance(stride, str) and stride == 'effective')):
            raise ValueError("Stride argument must either be an integer value or 'effective', "
                             "but was: {}".format(stride))
        self._stride = stride

    def compute_observation_likelihood(self, data: Union[np.ndarray, List[np.ndarray]]):
        r""" Computes the likelihood of observed data under this model.

        Internally, the forward pass of the Baum-Welch algorithm is used.

        Parameters
        ----------
        data : array_like or list of array_like
            The observations

        Returns
        -------
        likelihood : float
            The computed likelihood.
        """
        if not isinstance(data, (list, tuple)):
            data = [data]
        max_n_frames = max(len(obs) for obs in data)

        # get parameters
        A = self.transition_model.transition_matrix
        pi = self.initial_distribution
        alpha = np.zeros((max_n_frames, self.n_hidden_states), dtype=A.dtype)

        # compute output probability matrix
        loglik = 0.
        for obs in data:
            T = len(obs)
            pobs = self.output_model.to_state_probability_trajectory(obs)
            pobs = pobs.astype(A.dtype)
            loglik += forward_impl(A, pobs, pi, alpha_out=alpha, T=T)
        return loglik

    @property
    def lagtime(self) -> int:
        r""" The lagtime this model was estimated at.

        Returns
        -------
        lagtime : int
            The lagtime.
        """
        return self.transition_model.lagtime

    @property
    def stride(self):
        r"""
        The stride parameter which was used to subsample the discrete trajectories when estimating the hidden
        markov state model. Can either be an integer value or 'effective', in which case a stride is estimated at
        which subsequent states are uncorrelated.

        Returns
        -------
        stride : int or str
            The stride parameter.
        """
        return self._stride

    @property
    def observation_symbols(self) -> Optional[np.ndarray]:
        r"""
        The symbols represented by this HMM in observation space. Can be None in case the output model has no
        discrete observations it is None.

        Returns
        -------
        The list of observation symbols or None.
        """
        return self._observation_symbols

    @property
    def observation_symbols_full(self) -> Optional[np.ndarray]:
        r""" All symbols that the original model contained (original before taking any submodel).

        Returns
        -------
        The list of observation symbols or None, if there are no discrete symbols or None was provided.
        """
        return self._observation_symbols_full

    @property
    def n_observation_states(self) -> int:
        r"""
        Property determining the number of observed/macro states. It coincides with the size of the second axis
        of the observation probabilities matrix in case of a discrete output model.

        Returns
        -------
        Number of observed/macro states
        """
        return self.output_model.n_observable_states

    @property
    def count_model(self):
        r"""
        Yields the count model for the micro (hidden) states. The count matrix is estimated from Viterbi paths.

        Returns
        -------
        count_model : deeptime.markov.TransitionCountModel
            The count model for the micro states.
        """
        return self.transition_model.count_model

    @property
    def transition_model(self):
        r""" Yields the transition model for the hidden states.

        Returns
        -------
        model : deeptime.markov.msm.MarkovStateModel
            The transition model.
        """
        return self._transition_model

    @property
    def initial_count(self) -> Optional[np.ndarray]:
        r"""
        The hidden initial counts, can be None.

        Returns
        -------
        Initial counts.
        """
        return self._initial_count

    @property
    def output_model(self) -> OutputModel:
        r"""
        The selected output model for this HMM. The output model can map from the hidden states to observable states
        and can also be fitted to data.

        Returns
        -------
        The output model
        """
        return self._output_model

    @property
    def initial_distribution(self) -> np.ndarray:
        r""" The initial distribution of this HMM over the hidden states.

        Returns
        -------
        The initial distribution.
        """
        return self._initial_distribution

    @property
    def n_hidden_states(self) -> int:
        r"""
        The number of hidden states. Can also be retrieved from the output model as well as from the transition model.

        Returns
        -------
        Number of hidden states
        """
        return self.output_model.n_hidden_states

    @property
    def likelihoods(self) -> Optional[np.ndarray]:
        r"""
        If the model comes from the MaximumLikelihoodHMM estimator, this property contains the sequence of likelihoods
        generated from the fitting iteration.

        Returns
        -------
        Sequence of likelihoods, otherwise None.
        """
        return self._likelihoods

    @property
    def likelihood(self) -> Optional[float]:
        r"""
        The estimated likelihood of this model based on the training data. Only available if the sequence of likelihoods
        is provided.

        Returns
        -------
        The estimated likelihood, otherwise None.
        """
        return self.likelihoods[-1] if self.likelihoods is not None else None

    @property
    def state_probabilities(self) -> Optional[List[np.ndarray]]:
        r"""
        List of state probabilities for each trajectory that the model was trained on (gammas in the Baum-Welch algo).

        Returns
        -------
        List of state probabilities if initially provided in the constructor.
        """
        return self._state_probabilities

    @property
    def transition_counts(self) -> Optional[np.ndarray]:
        r"""
        The transition counts for the hidden states as estimated in the fitting procedure.

        Returns
        -------
        The transition counts, can be None if the transition model has no count model.
        """
        return self.transition_model.count_model.count_matrix if self.transition_model.count_model is not None else None

    @property
    def hidden_state_trajectories(self) -> Optional[List[np.ndarray]]:
        r"""
        Training trajectories mapped to hidden states after estimation.

        Returns
        -------
        hidden state trajectories, can be None if not provided in constructor.
        """
        return self._hidden_state_trajectories

    @property
    def output_probabilities(self) -> np.ndarray:
        r"""
        Returns the probabilities for each hidden state to map to a particular observation state. Only available
        if the underlying output model is a `DiscreteOutputModel`.

        Returns
        -------
        probabilities : np.ndarray
            a (M,N) row-stochastic matrix mapping from each hidden to each observation state
        """
        if isinstance(self.output_model, DiscreteOutputModel):
            return self.output_model.output_probabilities
        raise ValueError("Output probabilities are only available for HMMs with discrete output model.")

    @property
    def stationary_distribution_obs(self):
        r"""
        The stationary distribution in observable space. Only available with a discrete output model.

        Returns
        -------
        stationary distribution in observation space if available
        """
        if isinstance(self.output_model, DiscreteOutputModel):
            return np.dot(self.transition_model.stationary_distribution, self.output_probabilities)
        raise RuntimeError("only available for discrete output model")

    @property
    def eigenvectors_left_obs(self):
        r"""
        Left eigenvectors in observation space. Only available with a discrete output model.

        Returns
        -------
        Left eigenvectors in observation space.
        """
        return np.dot(self.transition_model.eigenvectors_left(), self.output_probabilities)

    @property
    def eigenvectors_right_obs(self):
        r"""
        Right eigenvectors in observation space. Only available with a discrete output model.

        Returns
        -------
        Right eigenvectors in observation space.
        """
        return np.dot(self.metastable_memberships, self.transition_model.eigenvectors_right())

    def transition_matrix_obs(self, k=1) -> np.ndarray:
        r""" Computes the transition matrix between observed states

        Transition matrices for longer lag times than the one used to parametrize this HMM can be obtained by setting
        the k option. Note that a HMM is not Markovian, thus we cannot compute transition matrices at longer lag times
        using the Chapman-Kolmogorow equality. I.e.:

        .. math::
            P (k \tau) \neq P^k (\tau)

        This function computes the correct transition matrix using the metastable (coarse)
        transition matrix :math:`P_c` as:

        .. math::
            P (k \tau) = {\Pi}^{-1} \chi^{\top} ({\Pi}_c) P_c^k (\tau) \chi

        where :math:`\chi` is the output probability matrix, :math:`\Pi_c` is a diagonal matrix with the
        metastable-state (coarse) stationary distribution and :math:`\Pi` is a diagonal matrix with the
        observable-state stationary distribution.

        Parameters
        ----------
        k : int, optional, default=1
            Multiple of the lag time.
            By default (k=1), the transition matrix at the lag time used to
            construct this HMM will be returned. If a higher power is given,

        """
        Pi_c = np.diag(self.transition_model.stationary_distribution)
        P_c = self.transition_model.transition_matrix
        P_c_k = np.linalg.matrix_power(P_c, k)  # take a power if needed
        B = self.output_probabilities
        C = np.dot(np.dot(B.T, Pi_c), np.dot(P_c_k, B))
        P = C / C.sum(axis=1)[:, None]  # row normalization
        return P

    @property
    def lifetimes(self) -> np.ndarray:
        r""" Lifetimes of states of the hidden transition matrix

        Returns
        -------
        l : ndarray(n_states)
            state lifetimes in units of the input trajectory time step,
            defined by :math:`-\tau / \ln \mid p_{ii} \mid, i = 1,...,n_\mathrm{states}`, where
            :math:`p_{ii}` are the diagonal entries of the hidden transition matrix.
        """
        return -self.transition_model.lagtime / np.log(np.diag(self.transition_model.transition_matrix))

    def compute_viterbi_paths(self, observations) -> List[np.ndarray]:
        r"""
        Computes the Viterbi paths using the current HMM model.

        Parameters
        ----------
        observations : list of array_like or array_like
            observations

        Returns
        -------
        paths : list of np.ndarray
            the computed viterbi paths
        """
        observations = ensure_dtraj_list(observations)
        A = self.transition_model.transition_matrix
        pi = self.initial_distribution
        state_probabilities = [
            self.output_model.to_state_probability_trajectory(obs) for obs in observations
        ]
        paths = [viterbi(A, obs, pi) for obs in state_probabilities]
        return paths

    def collect_observations_in_state(self, observations: List[np.ndarray], state_index: int):
        """Collect a vector of all observations belonging to a specified hidden state.

        Parameters
        ----------
        observations : list of numpy.array
            List of observed trajectories.
        state_index : int
            The index of the hidden state for which corresponding observations are to be retrieved.

        Returns
        -------
        collected_observations : numpy.array with shape (nsamples,)
            The collected vector of observations belonging to the specified hidden state.

        Raises
        ------
        RuntimeError
            A RuntimeError is raised if the HMM model does not yet have a hidden state trajectory associated with it.

        """
        if not self.hidden_state_trajectories:
            raise RuntimeError('HMM model does not have a hidden state trajectory.')
        observations = ensure_dtraj_list(observations)

        from ._util import observations_in_state
        return observations_in_state(self.hidden_state_trajectories, observations, state_index)

    ################################################################################
    # Generation of trajectories and samples
    ################################################################################

    def simulate(self, n_steps, start=None, stop=None, dt=1):
        """
        Generates a realization of the Hidden Markov Model

        Parameters
        ----------
        n_steps : int
            trajectory length in steps of the lag time
        start : int, optional, default = None
            starting hidden state. If not given, will sample from the stationary
            distribution of the hidden transition matrix.
        stop : int or int-array-like, optional, default = None
            stopping hidden set. If given, the trajectory will be stopped before
            N steps once a hidden state of the stop set is reached
        dt : int
            trajectory will be saved every dt time steps.
            Internally, the dt'th power of P is taken to ensure a more efficient simulation.

        Returns
        -------
        htraj : (N/dt, ) ndarray
            The hidden state trajectory with length N/dt
        otraj : (N/dt, ) ndarray
            The observable state discrete trajectory with length N/dt

        """
        # sample hidden trajectory
        htraj = self.transition_model.simulate(n_steps, start=start, stop=stop, dt=dt)
        otraj = self.output_model.generate_observation_trajectory(htraj)
        return htraj, otraj

    def transform_discrete_trajectories_to_observed_symbols(self, dtrajs):
        r"""A list of integer arrays with the discrete trajectories mapped to the currently used set of observation
        symbols. For example, if there has been a subselection of the model for connectivity='largest', the indices
        will be given within the connected set, frames that do not correspond to a considered symbol are set to -1.

        Parameters
        ----------
        dtrajs : array_like or list of array_like
            discretized trajectories

        Returns
        -------
        array_like or list of array_like
            Curated discretized trajectories so that unconsidered symbols are mapped to -1.
        """
        dtrajs = ensure_dtraj_list(dtrajs)

        max_state = self.observation_symbols_full.max()
        # add one because of 0 indexing and another one so that -1 gets mapped to -1
        mapping = np.full(max_state + 2, -1, dtype=np.int32)
        mapping[self.observation_symbols] = self.observation_symbols

        # map elements which are too large to -1 directly
        transformed_dtrajs = []
        for dtraj in dtrajs:
            transformed_dtrajs.append(dtraj.copy())
            transformed_dtrajs[-1][np.where(dtraj > max_state)[0]] = -1

        # perform mapping
        return [mapping[dtraj] for dtraj in transformed_dtrajs]

    def sample_by_observation_probabilities(self, dtrajs, nsample):
        r"""Generates samples according to the current observation probability distribution.

        Notes
        -----
        Sampling from off-sample-trajectories might yield -1 indices as discrete observable states
        are drawn from output probability distributions and off-sample trajectories might not
        contain all drawn observable states.

        Parameters
        ----------
        dtrajs : discrete trajectory
            Input observation trajectory or list of trajectories
        nsample : int
            Number of samples per distribution.

        Returns
        -------
        indexes : length m list of ndarray( (nsample, 2) )
            List of the sampled indices by distribution.
            Each element is an index array with a number of rows equal to nsample, with rows consisting of a
            tuple (i, t), where i is the index of the trajectory and t is the time index within the trajectory.
        """
        mapped = self.transform_discrete_trajectories_to_observed_symbols(dtrajs)
        if all(np.all(x == -1) for x in mapped):
            raise ValueError("The discrete trajectories contained no elements which are in the observation "
                             "symbols of this HMM.")
        observable_state_indices = sample.compute_index_states(mapped, subset=self.observation_symbols)
        return sample.indices_by_distribution(observable_state_indices, self.output_probabilities, nsample)

    # ================================================================================================================
    # Metastable state stuff based on HMM transition matrix
    # ================================================================================================================

    @property
    def metastable_memberships(self):
        r""" Computes the memberships of observable states to metastable sets by Bayesian inversion.
        :cite:`hmm-noe2013projected`

        Returns
        -------
        M : ndarray((n,m))
            A matrix containing the probability or membership of each
            observable state to be assigned to each metastable or hidden state.
            The row sums of M are 1.
        """
        nonzero = np.nonzero(self.stationary_distribution_obs)[0]
        M = np.zeros((self.n_observation_states, self.transition_model.n_states))
        M[nonzero, :] = np.transpose(np.diag(self.transition_model.stationary_distribution).dot(
            self.output_probabilities[:, nonzero]) / self.stationary_distribution_obs[nonzero])
        # renormalize
        M[nonzero, :] /= M.sum(axis=1)[nonzero, None]
        return M

    @property
    def metastable_distributions(self):
        r""" Returns the output probability distributions. Identical to
            :meth:`output_probabilities`

        Returns
        -------
        Pout : ndarray (m,n)
            output probability matrix from hidden to observable discrete states

        See also
        --------
        output_probabilities

        """
        return self.output_probabilities

    @property
    def metastable_sets(self):
        r""" Computes the metastable sets of observable states within each
            metastable set

        Notes
        -----
        This is only recommended for visualization purposes. You *cannot*
        compute any actual quantity of the coarse-grained kinetics without
        employing the fuzzy memberships!

        Returns
        -------
        sets : list of int-arrays
            A list of length equal to metastable states. Each element is an array
            with observable state indexes contained in it

        """
        res = []
        assignment = self.metastable_assignments
        for i in range(self.transition_model.n_states):
            res.append(np.where(assignment == i)[0])
        return res

    @property
    def metastable_assignments(self):
        r""" Computes the assignment to metastable sets for observable states

        Notes
        -----
        This is only recommended for visualization purposes. You *cannot*
        compute any actual quantity of the coarse-grained kinetics without
        employing the fuzzy memberships!

        Returns
        -------
        ndarray((n) ,dtype=int)
            For each observable state, the metastable state it is located in.

        See Also
        --------
        output_probabilities

        """
        return np.argmax(self.output_probabilities, axis=0)

    # ================================================================================================================
    # Micro- / observable state properties
    # ================================================================================================================

    def _project_to_hidden(self, a: np.ndarray, ndim=None, allow_none=None) -> Optional[np.ndarray]:
        r"""
        Projects an observable state vector to hidden states. Only available for discrete output model.

        Parameters
        ----------
        a : array_like
            observable state vector
        ndim : int, optional, default=None
            requested dimension of observable state vector
        allow_none : bool, optional, default=None
            Whether None is allowed as observable state vector, resulting in a no-op.
        Returns
        -------
        hidden state vector corresponding to observable state vector
        """
        if allow_none and a is None:
            return None
        a = ensure_array(a, ndim=ndim)
        if len(a) != self.n_observation_states:
            raise ValueError("Input array has incompatible shape, needs to have "
                             "length {} but had length {}.".format(len(a), self.n_observation_states))
        return np.dot(self.output_probabilities, a)

    def expectation_obs(self, a):
        r"""Equilibrium expectation value of a given observable state vector.

        See Also
        --------
        deeptime.markov.msm.MarkovStateModel.expectation
        """
        return self.transition_model.expectation(self._project_to_hidden(a))

    def correlation_obs(self, a, b=None, maxtime=None, k=None, ncv=None):
        r""" Time-correlation for equilibrium experiment based on observable state vectors a and b.

        See Also
        --------
        deeptime.markov.msm.MarkovStateModel.correlation
        """
        # basic checks for a and b
        a = self._project_to_hidden(a)
        b = self._project_to_hidden(b, ndim=1, allow_none=True)
        return self.transition_model.correlation(a, b=b, maxtime=maxtime, k=k, ncv=ncv)

    def fingerprint_correlation_obs(self, a, b=None, k=None, ncv=None):
        r""" Dynamical fingerprint for equilibrium time-correlation experiment based on observable state vectors a
        and b.

        See Also
        --------
        deeptime.markov.msm.MarkovStateModel.fingerprint_correlation
        """
        a = self._project_to_hidden(a, ndim=1)
        b = self._project_to_hidden(b, ndim=1, allow_none=True)
        return self.transition_model.fingerprint_correlation(a, b=b, k=k, ncv=ncv)

    def relaxation_obs(self, p0, a, maxtime=None, k=None, ncv=None):
        r""" Simulates a perturbation-relaxation experiment based on observable state vector and distribution.

        See Also
        --------
        deeptime.markov.msm.MarkovStateModel.relaxation
        """
        p0 = self._project_to_hidden(p0, ndim=1)
        a = self._project_to_hidden(a, ndim=1)
        return self.transition_model.relaxation(p0, a, maxtime=maxtime, k=k, ncv=ncv)

    def fingerprint_relaxation_obs(self, p0, a, k=None, ncv=None):
        r""" Dynamical fingerprint for perturbation/relaxation experiment
        based on observable state vector and distribution.

        See Also
        --------
        deeptime.markov.msm.MarkovStateModel.fingerprint_relaxation
        """
        p0 = self._project_to_hidden(p0, ndim=1)
        a = self._project_to_hidden(a, ndim=1)
        return self.transition_model.fingerprint_relaxation(p0, a, k=k, ncv=ncv)

    def propagate(self, p0, k):
        r""" Propagates the initial distribution p0 defined on observable space k times.

        Therefore computes the product

        .. math::

            p_k = p_0^T P^k

        If the lag time of transition matrix :math:`P` is :math:`\tau`, this
        will provide the probability distribution at time :math:`k \tau`.

        Parameters
        ----------
        p0 : ndarray(n)
            Initial distribution. Vector of size of the active set.

        k : int
            Number of time steps

        Returns
        ----------
        pk : ndarray(n)
            Distribution after k steps

        """
        if k == 0:  # simply return p0 normalized
            return p0 / p0.sum()

        p0 = self._project_to_hidden(p0)

        ev_right = self.transition_model.eigenvectors_right(self.n_hidden_states)
        ev_left = self.transition_model.eigenvectors_left(self.n_hidden_states)
        pk = np.linalg.multi_dot([p0.T, ev_right, np.diag(np.power(self.transition_model.eigenvalues(), k)), ev_left])

        pk = np.dot(pk, self.output_probabilities)  # convert back to microstate space

        # normalize to 1.0 and return
        return pk / pk.sum()

    # ================================================================================================================
    # Submodel functionality
    # ================================================================================================================

    def submodel(self, states: Optional[np.ndarray] = None, obs: Optional[np.ndarray] = None):
        """Returns a HMM with restricted state space

        Parameters
        ----------
        states : None or int-array
            Hidden states to restrict the model to. In addition to specifying
            the subset, possible options are:

            * int-array: indices of states to restrict onto
            * None : all states - don't restrict

        obs : None or int-array
            Observed states to restrict the model to. In addition to specifying
            an array with the state labels to be observed, possible options are:

              * int-array: indices of states to restrict onto
              * None : all states - don't restrict
        Returns
        -------
        hmm : HiddenMarkovModel
            The restricted HMM.
        """

        if states is None and obs is None:
            return self  # do nothing
        if states is None:
            states = np.arange(self.n_hidden_states)
        if obs is None:
            obs = np.arange(self.n_observation_states)

        transition_model = self.transition_model.submodel(states)

        if self.initial_count is not None:
            initial_count = self.initial_count[states].copy()
        else:
            initial_count = None
        initial_distribution = self.initial_distribution[states] / np.sum(self.initial_distribution[states])

        # observation matrix
        output_model = self.output_model.submodel(states, obs)

        observation_symbols = self.observation_symbols
        if observation_symbols is not None:
            observation_symbols = observation_symbols[obs]

        model = HiddenMarkovModel(transition_model=transition_model, output_model=output_model,
                                  initial_distribution=initial_distribution, likelihoods=self.likelihoods,
                                  state_probabilities=self.state_probabilities, initial_count=initial_count,
                                  hidden_state_trajectories=self.hidden_state_trajectories,
                                  observation_symbols=observation_symbols,
                                  observation_symbols_full=self.observation_symbols_full)
        return model

    def _select_states(self, connectivity_threshold, states) -> np.ndarray:
        r"""
        Retrieves a selection of states based on the arguments provided.

        Parameters
        ----------
        connectivity_threshold : str or int
            A connectivity threshold that needs to be exceeded so that an edge in the connectivity graph is considered
            to be relevant. In case of a string it can be set to "1/n", where n is the number of hidden states.
        states : array_like
            The states to restrict onto. Can be further reduced by providing a threshold.
        Returns
        -------
        A ndarray containing a subselection of states.
        """
        if str(connectivity_threshold) == '1/n':
            connectivity_threshold = 1.0 / float(self.transition_model.n_states)
        if isinstance(states, str):
            strong = 'strong' in states
            largest = 'largest' in states
            count_model = self.transition_model.count_model
            S = count_model.connected_sets(connectivity_threshold=connectivity_threshold, directed=strong)
            if largest:
                score = np.array([len(s) for s in S])
            else:
                score = np.array([count_model.count_matrix[np.ix_(s, s)].sum() for s in S])
            states = S[np.argmax(score)]
        return states

    def nonempty_obs(self, dtrajs) -> np.ndarray:
        r"""
        Computes the set of visited observable states given a set of discrete trajectories.

        Parameters
        ----------
        dtrajs : array_like
            observable trajectory

        Returns
        -------
        symbols : np.ndarray
            The observation symbols which are visited.
        """
        from deeptime.markov import compute_dtrajs_effective, count_states
        if dtrajs is None:
            raise ValueError("Needs nonempty dtrajs to evaluate nonempty obs.")
        dtrajs = ensure_dtraj_list(dtrajs)
        dtrajs_lagged_strided = compute_dtrajs_effective(
            dtrajs, self.transition_model.lagtime, self.transition_model.count_model.n_states_full, self.stride
        )
        obs = np.where(count_states(dtrajs_lagged_strided) > 0)[0]
        return obs

    def states_largest(self, directed=True, connectivity_threshold='1/n') -> np.ndarray:
        r"""
        Selects hidden states which represent the largest connected set.

        Parameters
        ----------
        directed : bool, optional, default=True
            Whether the connectivity is strong (directed) or weak (undirected)
        connectivity_threshold : str or int, optional, default='1/n'
            A connectivity threshold which can be employed to only consider edges with a certain minimum weight.

        Returns
        -------
        The largest connected set of hidden states
        """
        return self._select_states(connectivity_threshold, 'largest-strong' if directed else 'largest-weak')

    def submodel_largest(self, directed=True, connectivity_threshold='1/n', observe_nonempty=True, dtrajs=None):
        r"""
        Returns the largest connected sub-HMM. By default this means that the largest connected set of hidden states
        and the set of visited observable states is selected.

        Parameters
        ----------
        directed : bool, optional, default=True
            Whether the connectivity is based on a directed graph (strong connectiviy) or undirected (weak connectivity)
        connectivity_threshold : str or int, optional, default='1/n'
            The connectivity threshold required to consider two hidden states connected.
        observe_nonempty : bool, optional, default=True
            Whether the observable state set should be restricted to visited observable states. If True, dtrajs must
            be provided.
        dtrajs : array_like, optional, default=None
            Observable state trajectory or a list thereof to evaluate visited observable states.

        Returns
        -------
        sub_hmm : HiddenMarkovModel
            The restricted HMM.
        """
        states = self.states_largest(directed=directed, connectivity_threshold=connectivity_threshold)
        obs = self.nonempty_obs(dtrajs) if observe_nonempty else None
        return self.submodel(states=states, obs=obs)

    def states_populous(self, strong=True, connectivity_threshold='1/n'):
        r"""
        Retrieves the hidden states which are most populated and connected.

        Parameters
        ----------
        strong : bool, optional, default=True
            Whether the connectivity is evaluated based on a directed or on an undirected graph.
        connectivity_threshold : str or int, optional, default=None
            Minimum weight so that two states are considered connected.

        Returns
        -------
        states : np.ndarray
            Most populated set of states
        """
        return self._select_states(connectivity_threshold, 'populous-strong' if strong else 'populous-weak')

    def submodel_populous(self, directed=True, connectivity_threshold='1/n', observe_nonempty=True, dtrajs=None):
        """ Returns the most populous connected sub-HMM.

        Parameters
        ----------
        directed : bool, optional, default=True
            Whether the connectivity is based on a directed graph (strong connectiviy) or undirected (weak connectivity)
        connectivity_threshold : str or int, optional, default='1/n'
            The connectivity threshold required to consider two hidden states connected.
        observe_nonempty : bool, optional, default=True
            Whether the observable state set should be restricted to visited observable states. If True, dtrajs must
            be provided.
        dtrajs : array_like, optional, default=None
            Observable state trajectory or a list thereof to evaluate visited observable states.

        Returns
        -------
        hmm : HiddenMarkovModel
            The restricted HMM.

        """
        states = self.states_populous(strong=directed, connectivity_threshold=connectivity_threshold)
        obs = self.nonempty_obs(dtrajs) if observe_nonempty else None
        return self.submodel(states=states, obs=obs)

    def submodel_disconnect(self, connectivity_threshold='1/n'):
        """Disconnects sets of hidden states that are barely connected

        Runs a connectivity check excluding all transition counts below
        connectivity_threshold. The transition matrix and stationary distribution
        will be re-estimated. Note that the resulting transition matrix
        may have both strongly and weakly connected subsets.

        Parameters
        ----------
        connectivity_threshold : float or '1/n'
            minimum number of counts to consider a connection between two states.
            Counts lower than that will count zero in the connectivity check and
            may thus separate the resulting transition matrix. The default
            evaluates to 1/n_states.

        Returns
        -------
        hmm : HiddenMarkovModel
            The restricted HMM.
        """
        lcc = self.transition_model.count_model.connected_sets(connectivity_threshold=connectivity_threshold)[0]
        return self.submodel(lcc)

    # ================================================================================================================
    # Properties inherited from transition model
    # ================================================================================================================


def viterbi(transition_matrix: np.ndarray, state_probability_trajectory: np.ndarray, initial_distribution: np.ndarray):
    """ Estimate the hidden pathway of maximum likelihood using the Viterbi algorithm.

    Parameters
    ----------
    transition_matrix : ndarray((N,N), dtype = float)
        transition matrix of the hidden states
    state_probability_trajectory : ndarray((T,N), dtype = float)
        pobs[t,i] is the observation probability for observation at time t given hidden state i
    initial_distribution : ndarray((N), dtype = float)
        initial distribution of hidden states

    Returns
    -------
    q : numpy.array shape (T)
        maximum likelihood hidden path

    """
    if state_probability_trajectory.ndim == 1 and transition_matrix.shape[0] == 1:
        # if there is only one state, pad so that there is an additional dimension
        state_probability_trajectory = state_probability_trajectory[..., None]
    return viterbi_impl(transition_matrix=transition_matrix,
                        state_probability_trajectory=state_probability_trajectory,
                        initial_distribution=initial_distribution)



from typing import Optional, Union, List, Callable

import numpy as np

from . import vamp_score
from ..base import Model, Transformer, Estimator
from ..basis import Identity, Concatenation
from ..covariance import CovarianceModel, WhiteningTransform
from ..numeric import is_diagonal_matrix
from ..util import LaggedModelValidator
from ..util.decorators import cached_property


class TransferOperatorModel(Model, Transformer):
    r""" Model which contains a finite-dimensional transfer operator (or approximation thereof).
    It describes the temporal evolution of observable space, i.e.,

    .. math:: \mathbb{E}[g(x_{t+\tau})] = K^\top \mathbb{E}[f(x_t)],

    where :math:`K\in\mathbb{R}^{n\times m}` is the transfer operator, :math:`x_t` the system's state at time :math:`t`,
    and :math:`f` and :math:`g` observables of the system's state.

    Parameters
    ----------
    koopman_matrix : (n, n) ndarray
        Applies the transform :math:`K^\top` in the modified basis.
    instantaneous_obs : Callable, optional, default=identity
        Transforms the current state :math:`x_t` to :math:`f(x_t)`. Defaults to `f(x) = x`.
    timelagged_obs : Callable, optional, default=identity
        Transforms the future state :math:`x_{t+\tau}` to :math:`g(x_{t+\tau})`. Defaults to `f(x) = x`.
    """

    def __init__(self, koopman_matrix: np.ndarray,
                 instantaneous_obs: Callable[[np.ndarray], np.ndarray] = Identity(),
                 timelagged_obs: Callable[[np.ndarray], np.ndarray] = Identity()):
        super().__init__()
        self._koopman_matrix = koopman_matrix
        self._instantaneous_obs = instantaneous_obs
        self._timelagged_obs = timelagged_obs

    @property
    def operator(self) -> np.ndarray:
        r""" The operator :math:`K` so that :math:`\mathbb{E}[g(x_{t+\tau}] = K^\top \mathbb{E}[f(x_t)]` in transformed
        bases.
        """
        return self._koopman_matrix

    @property
    def koopman_matrix(self) -> np.ndarray:
        r""" Same as :attr:`operator`. """
        return self.operator

    @cached_property
    def operator_inverse(self) -> np.ndarray:
        r""" Inverse of the operator :math:`K`, i.e., :math:`K^{-1}`.
        Potentially pseudo-inverse instead of true inverse.
        """
        return np.linalg.pinv(self.operator)

    @property
    def instantaneous_obs(self) -> Callable[[np.ndarray], np.ndarray]:
        r""" Transforms the current state :math:`x_t` to :math:`f(x_t)`. Defaults to `f(x) = x`. """
        return self._instantaneous_obs

    @property
    def timelagged_obs(self) -> Callable[[np.ndarray], np.ndarray]:
        r""" Transforms the future state :math:`x_{t+\tau}` to :math:`g(x_{t+\tau})`. Defaults to `f(x) = x`. """
        return self._timelagged_obs

    @property
    def output_dimension(self):
        r""" The dimension of data after propagation by :math:`K`. """
        return self.operator.shape[1]

    def forward(self, data: np.ndarray, propagate=True):
        r""" Maps data forward in time.

        Parameters
        ----------
        data : (T, n) ndarray
            Input data
        propagate : bool, default=True
            Whether to apply the Koopman operator to the featurized data.

        Returns
        -------
        mapped_data : (T, m) ndarray
            Mapped data.
        """
        out = self.instantaneous_obs(data)
        if propagate:
            out = out @ self.operator
        return out

    def backward(self, data: np.ndarray, propagate=True):
        r""" Maps data backward in time.

        Parameters
        ----------
        data : (T, n) ndarray
            Input data
        propagate : bool, default=True
            Whether to apply the Koopman operator to the featurized data.

        Returns
        -------
        mapped_data : (T, m) ndarray
            Mapped data.
        """
        out = self.timelagged_obs(data)
        if propagate:
            out = out @ self.operator_inverse
        return out

    def transform(self, data: np.ndarray, **kw):
        r""" Transforms data by applying the observables to it and then mapping them onto the modes of :math:`K`.

        Parameters
        ----------
        data : (T,n) ndarray
            Input data.

        Returns
        -------
        transformed_data : (T,m) ndarray
            The transformed data.
        """
        return self.instantaneous_obs(data)


class CovarianceKoopmanModel(TransferOperatorModel):
    r"""A type of Koopman model :math:`\mathbb{E}[g(x_{t+\tau})] = K^\top \mathbb{E}[f(x_{t})]`
    which was obtained through diagonalization of covariance matrices. This leads to
    a Koopman operator which is a diagonal matrix and can be used to project onto specific processes of the system.

    In particular, this model expects matrices :math:`U` and :math:`V` as well as singular values :math:`\sigma_i`,
    such that

    .. math::
        \mathbb{E}[V^\top\chi_1 (x_{t+\tau})]=\mathbb{E}[g(x_{t+\tau})] \approx K^\top \mathbb{E}[f(x_{t})] = \mathrm{diag}(\sigma_i) \mathbb{E}[U^\top\chi_0(x_{t})],

    where :math:`\chi_0,\chi_1` are basis transformations of the full state :math:`x_t`.

    The estimators which produce this kind of model are :class:`VAMP <deeptime.decomposition.VAMP>` and
    :class:`TICA <deeptime.decomposition.TICA>`.

    For a description of parameters `operator`, `basis_transform_forward`, `basis_transform_backward`,
    and `output_dimension`: please see :class:`TransferOperatorModel`.

    Parameters
    ----------
    instantaneous_coefficients : (n, k) ndarray
        The coefficient matrix :math:`U`.
    singular_values : (k,) ndarray
        Singular values :math:`\sigma_i`.
    instantaneous_coefficients : (n, k) ndarray
        The coefficient matrix :math:`V`.
    cov : CovarianceModel
        Covariances :math:`C_{00}`, :math:`C_{0t}`, and :math:`C_{tt}`.
    rank_0 : int
        Rank of the instantaneous whitening transformation :math:`C_{00}^{-1/2}`.
    rank_t : int
        Rank of the time-lagged whitening transformation :math:`C_{tt}^{-1/2}`.
    scaling : str or None, default=None
        Scaling parameter which was applied to singular values for additional structure in the projected space.
        See the respective estimator for details.
    epsilon : float, default=1e-6
        Eigenvalue / singular value cutoff. Eigenvalues (or singular values) of :math:`C_{00}` and :math:`C_{11}`
        with norms <= epsilon were cut off. The remaining number of eigenvalues together with the value of `dim`
        define the effective output dimension.
    instantaneous_obs : Callable, optional, default=identity
        Transforms the current state :math:`x_t` to :math:`\chi_0(x_t)`. Defaults to `\chi_0(x) = x`.
    timelagged_obs : Callable, optional, default=identity
        Transforms the future state :math:`x_{t+\tau}` to :math:`\chi_1(x_{t+\tau})`. Defaults to `\chi_1(x) = x`.
    """

    def __init__(self, instantaneous_coefficients, singular_values, timelagged_coefficients, cov,
                 rank_0: int, rank_t: int, dim=None, var_cutoff=None, scaling=None, epsilon=1e-10,
                 instantaneous_obs: Callable[[np.ndarray], np.ndarray] = Identity(),
                 timelagged_obs: Callable[[np.ndarray], np.ndarray] = Identity()):
        if not singular_values.ndim == 1:
            assert is_diagonal_matrix(singular_values)
            singular_values = np.diag(singular_values)
        self._whitening_instantaneous = WhiteningTransform(instantaneous_coefficients,
                                                           cov.mean_0 if cov.data_mean_removed else None)
        self._whitening_timelagged = WhiteningTransform(timelagged_coefficients,
                                                        cov.mean_t if cov.data_mean_removed else None)
        super().__init__(np.diag(singular_values),
                         Concatenation(self._whitening_instantaneous, instantaneous_obs),
                         Concatenation(self._whitening_timelagged, timelagged_obs))
        self._instantaneous_coefficients = instantaneous_coefficients
        self._timelagged_coefficients = timelagged_coefficients
        self._singular_values = singular_values
        self._cov = cov

        self._scaling = scaling
        self._epsilon = epsilon
        self._rank_0 = rank_0
        self._rank_t = rank_t
        self._dim = dim
        self._var_cutoff = var_cutoff
        self._update_output_dimension()

    @property
    def output_dimension(self):
        return self._output_dimension

    @property
    def instantaneous_coefficients(self) -> np.ndarray:
        r""" Coefficient matrix :math:`U`. """
        return self._instantaneous_coefficients

    @property
    def timelagged_coefficients(self) -> np.ndarray:
        r""" Coefficient matrix :math:`V`. """
        return self._timelagged_coefficients

    @property
    def scaling(self) -> Optional[str]:
        """Scaling of projection. Can be :code:`None`, 'kinetic map', or 'km' """
        return self._scaling

    @property
    def singular_vectors_left(self) -> np.ndarray:
        """Transformation matrix that represents the linear map from mean-free feature space
        to the space of left singular functions."""
        return self.instantaneous_coefficients

    @property
    def singular_vectors_right(self) -> np.ndarray:
        """Transformation matrix that represents the linear map from mean-free feature space
        to the space of right singular functions."""
        return self.timelagged_coefficients

    @property
    def singular_values(self) -> np.ndarray:
        """ The singular values of the half-weighted Koopman matrix. """
        return np.diag(self.operator)

    @property
    def cov(self) -> CovarianceModel:
        r""" Estimated covariances. """
        return self._cov

    @property
    def mean_0(self) -> np.ndarray:
        r""" Shortcut to :attr:`mean_0 <deeptime.covariance.CovarianceModel.mean_0>`. """
        return self.cov.mean_0

    @property
    def mean_t(self) -> np.ndarray:
        r""" Shortcut to :attr:`mean_t <deeptime.covariance.CovarianceModel.mean_t>`. """
        return self.cov.mean_t

    @property
    def cov_00(self) -> np.ndarray:
        r""" Shortcut to :attr:`cov_00 <deeptime.covariance.CovarianceModel.cov_00>`. """
        return self.cov.cov_00

    @property
    def cov_0t(self) -> np.ndarray:
        r""" Shortcut to :attr:`cov_0t <deeptime.covariance.CovarianceModel.cov_0t>`. """
        return self.cov.cov_0t

    @property
    def cov_tt(self) -> np.ndarray:
        r""" Shortcut to :attr:`cov_tt <deeptime.covariance.CovarianceModel.cov_tt>`. """
        return self.cov.cov_tt

    @property
    def whitening_rank_0(self) -> int:
        r""" Rank of the instantaneous whitening transformation :math:`C_{00}^{-1/2}`. """
        return self._rank_0

    @property
    def whitening_rank_t(self) -> int:
        r""" Rank of the time-lagged whitening transformation :math:`C_{tt}^{-1/2}`. """
        return self._rank_t

    @property
    def epsilon(self) -> float:
        r""" Singular value cutoff. """
        return self._epsilon

    @staticmethod
    def _cumvar(singular_values) -> np.ndarray:
        cumvar = np.cumsum(singular_values ** 2)
        cumvar /= cumvar[-1]
        return cumvar

    @staticmethod
    def effective_output_dimension(rank0, rankt, dim, var_cutoff, singular_values) -> int:
        r""" Computes effective output dimension. """
        if (dim is None and var_cutoff is None) or (var_cutoff is not None and var_cutoff == 1.0):
            return min(rank0, rankt)
        if var_cutoff is not None:
            return np.searchsorted(CovarianceKoopmanModel._cumvar(singular_values), var_cutoff) + 1
        else:
            if dim is None:
                return min(rank0, rankt)
            else:
                return np.min([rank0, rankt, dim])

    def _update_output_dimension(self):
        self._output_dimension = CovarianceKoopmanModel.effective_output_dimension(
            self.whitening_rank_0, self.whitening_rank_t, self.dim, self.var_cutoff, self.singular_values
        )
        self._whitening_instantaneous.dim = self.output_dimension
        self._whitening_timelagged.dim = self.output_dimension

    def backward(self, data: np.ndarray, propagate=True):
        out = self.timelagged_obs(data)
        if propagate:
            op = self.operator_inverse[:self.output_dimension, :self.output_dimension]
            out = out @ op
        return out

    def forward(self, data: np.ndarray, propagate=True):
        out = self.instantaneous_obs(data)
        if propagate:
            op = self.operator[:self.output_dimension, :self.output_dimension]
            out = out @ op
        return out

    def transform(self, data: np.ndarray, **kw):
        r"""Projects data onto the Koopman modes :math:`f(x) = U^\top \chi_0 (x)`, where :math:`U` are the
        coefficients of the basis :math:`\chi_0`.

        Parameters
        ----------
        data : (T, n) ndarray
            Input data.

        Returns
        -------
        transformed_data : (T, k) ndarray
            Data projected onto the Koopman modes.
        """
        return self.instantaneous_obs(data)

    @property
    def var_cutoff(self) -> Optional[float]:
        r""" Variance cutoff parameter. Can be set to include dimensions up to a certain threshold. Takes
        precedence over the :meth:`dim` parameter.

        :getter: Yields the current variance cutoff.
        :setter: Sets a new variance cutoff
        :type: float or None
        """
        return self._var_cutoff

    @var_cutoff.setter
    def var_cutoff(self, value):
        assert 0 < value <= 1., "Invalid dimension parameter, if it is given in terms of a variance cutoff, " \
                                "it can only be in the interval (0, 1]."
        self._var_cutoff = value
        self._update_output_dimension()

    @property
    def dim(self) -> Optional[int]:
        r""" Dimension attribute. Can either be int or None. In case of

        * :code:`int` it evaluates it as the actual dimension, must be strictly greater 0,
        * :code:`None` all numerically available components are used.

        :getter: yields the dimension
        :setter: sets a new dimension
        :type: int or None
        """
        return self._dim

    @dim.setter
    def dim(self, value: Optional[int]):
        if isinstance(value, int) and value <= 0:
            # first test against Integral as `isinstance(1, Real)` also evaluates to True
            raise ValueError("VAMP: Invalid dimension parameter, if it is given in terms of the "
                             "dimension (integer), must be positive.")
        self._dim = value
        self._update_output_dimension()

    @property
    def cumulative_kinetic_variance(self) -> np.ndarray:
        r""" Yields the cumulative kinetic variance. """
        return CovarianceKoopmanModel._cumvar(self.singular_values)

    @cached_property
    def _instantaneous_whitening_backwards(self):
        inv = np.linalg.pinv(self._whitening_instantaneous.sqrt_inv_cov)
        mean = self._whitening_instantaneous.mean
        if mean is None:
            mean = 0
        return lambda x: x @ inv + mean

    def propagate(self, trajectory: np.ndarray, components: Optional[Union[int, List[int]]] = None) -> np.ndarray:
        r""" Applies the forward transform to the trajectory in non-transformed space. Given the Koopman operator
        :math:`\Sigma`, transformations  :math:`V^\top - \mu_t` and :math:`U^\top -\mu_0` for
        bases :math:`f` and :math:`g`, respectively, this is achieved by transforming each frame :math:`X_t` with

        .. math::
            \hat{X}_{t+\tau} = (V^\top)^{-1} \Sigma U^\top (X_t - \mu_0) + \mu_t.

        If the model stems from a :class:`VAMP <deeptime.decomposition.VAMP>` estimator, :math:`V` are the left
        singular vectors, :math:`\Sigma` the singular values, and :math:`U` the right singular vectors.

        Parameters
        ----------
        trajectory : (T, n) ndarray
            The input trajectory
        components : int or list of int or None, default=None
            Optional arguments for the Koopman operator if appropriate. If the model stems from
            a :class:`VAMP <deeptime.decomposition.VAMP>` estimator, these are the component(s) to project onto.
            If None, all processes are taken into account, if list of integer, this sets all singular values
            to zero but the "components"th ones.

        Returns
        -------
        predictions : (T, n) ndarray
            The predicted trajectory.
        """
        if components is not None:
            operator = np.zeros_like(self.operator)
            if not isinstance(components, (list, tuple)):
                components = [components]
            for ii in components:
                operator[ii, ii] = self.operator[ii, ii]
        else:
            operator = self.operator
        out = self.instantaneous_obs(trajectory)  # (T, N) -> (T, n) project into whitened space
        out = out @ operator  # (T, n) -> (T, n) propagation in whitened space
        out = self._instantaneous_whitening_backwards(out)  # (T, n) -> (T, N) back into original space
        return out

    def score(self, r: Union[float, str], test_model=None, epsilon=1e-6, dim=None):
        """Compute the VAMP score between a this model and potentially a test model for cross-validation.

        Parameters
        ----------
        r : float or str
            The type of score to evaluate. Can by an floating point value greater or equal to 1 or 'E', yielding the
            VAMP-r score or the VAMP-E score, respectively. :footcite:`wu2020variational`
            Typical choices are:

            *  'VAMP1'  Sum of singular values of the half-weighted Koopman matrix.
                        If the model is reversible, this is equal to the sum of
                        Koopman matrix eigenvalues, also called Rayleigh quotient :footcite:`wu2020variational`.
            *  'VAMP2'  Sum of squared singular values of the half-weighted Koopman
                        matrix :footcite:`wu2020variational`. If the model is reversible, this is
                        equal to the kinetic variance :footcite:`noe2015kinetic`.
            *  'VAMPE'  Approximation error of the estimated Koopman operator with respect to
                        the true Koopman operator up to an additive constant :footcite:`wu2020variational` .

        test_model : CovarianceKoopmanModel, optional, default=None

            If `test_model` is not None, this method computes the cross-validation score
            between self and `covariances_test`. It is assumed that self was estimated from
            the "training" data and `test_model` was estimated from the "test" data. The
            score is computed for one realization of self and `test_model`. Estimation
            of the average cross-validation score and partitioning of data into test and
            training part is not performed by this method.

            If `covariances_test` is None, this method computes the VAMP score for the model
            contained in self.

        epsilon : float, default=1e-6
            Regularization parameter for computing sqrt-inverses of spd matrices.
        dim : int, optional, default=None
            How many components to use for scoring.

        Returns
        -------
        score : float
            If `test_model` is not None, returns the cross-validation VAMP score between
            self and `test_model`. Otherwise return the selected VAMP-score of self.

        Notes
        -----
        The VAMP-:math:`r` and VAMP-E scores are computed according to :footcite:`wu2020variational`,
        Equation (33) and Equation (30), respectively.

        References
        ----------
        .. footbibliography::
        """
        test_cov = test_model.cov if test_model is not None else None
        dim = self.output_dimension if dim is None else dim
        return vamp_score(self, r, test_cov, dim, epsilon)

    def expectation(self, observables, statistics, lag_multiple=1, observables_mean_free=False,
                    statistics_mean_free=False):
        r"""Compute future expectation of observable or covariance using the approximated Koopman operator.

        Parameters
        ----------
        observables : np.ndarray((input_dimension, n_observables))
            Coefficients that express one or multiple observables in
            the basis of the input features.

        statistics : np.ndarray((input_dimension, n_statistics)), optional
            Coefficients that express one or multiple statistics in
            the basis of the input features.
            This parameter can be None. In that case, this method
            returns the future expectation value of the observable(s).

        lag_multiple : int
            If > 1, extrapolate to a multiple of the estimator's lag
            time by assuming Markovianity of the approximated Koopman
            operator.

        observables_mean_free : bool, default=False
            If true, coefficients in `observables` refer to the input
            features with feature means removed.
            If false, coefficients in `observables` refer to the
            unmodified input features.

        statistics_mean_free : bool, default=False
            If true, coefficients in `statistics` refer to the input
            features with feature means removed.
            If false, coefficients in `statistics` refer to the
            unmodified input features.

        Returns
        -------
        expectation : ndarray
            The equilibrium expectation of observables or covariance if statistics is not None.

        Notes
        -----
        A "future expectation" of a observable g is the average of g computed
        over a time window that has the same total length as the input data
        from which the Koopman operator was estimated but is shifted
        by ``lag_multiple*tau`` time steps into the future (where tau is the lag
        time).

        It is computed with the equation:

        .. math::

            \mathbb{E}[g]_{\rho_{n}}=\mathbf{q}^{T}\mathbf{P}^{n-1}\mathbf{e}_{1}

        where

        .. math::

            P_{ij}=\sigma_{i}\langle\psi_{i},\phi_{j}\rangle_{\rho_{1}}

        and

        .. math::

            q_{i}=\langle g,\phi_{i}\rangle_{\rho_{1}}

        and :math:`\mathbf{e}_{1}` is the first canonical unit vector.


        A model prediction of time-lagged covariances between the
        observable f and the statistic g at a lag-time of ``lag_multiple*tau``
        is computed with the equation:

        .. math::

            \mathrm{cov}[g,\,f;n\tau]=\mathbf{q}^{T}\mathbf{P}^{n-1}\boldsymbol{\Sigma}\mathbf{r}

        where :math:`r_{i}=\langle\psi_{i},f\rangle_{\rho_{0}}` and
        :math:`\boldsymbol{\Sigma}=\mathrm{diag(\boldsymbol{\sigma})}` .
        """
        if lag_multiple <= 0:
            raise ValueError("lag_multiple <= 0 not implemented")

        dim = self.output_dimension

        S = np.diag(np.concatenate(([1.0], self.singular_values[0:dim])))
        V = self.singular_vectors_right[:, 0:dim]
        U = self.singular_vectors_left[:, 0:dim]
        m_0 = self.mean_0
        m_t = self.mean_t

        if lag_multiple == 1:
            P = S
        else:
            p = np.zeros((dim + 1, dim + 1))
            p[0, 0] = 1.0
            p[1:, 0] = U.T.dot(m_t - m_0)
            p[1:, 1:] = U.T.dot(self.cov_tt).dot(V)
            P = np.linalg.matrix_power(S.dot(p), lag_multiple - 1).dot(S)

        Q = np.zeros((observables.shape[1], dim + 1))
        if not observables_mean_free:
            Q[:, 0] = observables.T.dot(m_t)
        Q[:, 1:] = observables.T.dot(self.cov_tt).dot(V)

        if statistics is not None:
            # compute covariance
            R = np.zeros((statistics.shape[1], dim + 1))
            if not statistics_mean_free:
                R[:, 0] = statistics.T.dot(m_0)
            R[:, 1:] = statistics.T.dot(self.cov_00).dot(U)

            # compute lagged covariance
            return Q.dot(P).dot(R.T)
            # TODO: discuss whether we want to return this or the transpose
            # TODO: from MSMs one might expect to first index to refer to the statistics, here it is the other way round
        else:
            # compute future expectation
            return Q.dot(P)[:, 0]

    def timescales(self, k=None, lagtime: Optional[int] = None) -> np.ndarray:
        r"""Implied timescales of the TICA transformation

        For each :math:`i`-th eigenvalue, this returns

        .. math::

            t_i = -\frac{\tau}{\log(|\lambda_i|)}

        where :math:`\tau` is the :attr:`lagtime` of the TICA object and :math:`\lambda_i` is the `i`-th
        :attr:`eigenvalue <eigenvalues>` of the TICA object.

        Parameters
        ----------
        k : int, optional, default=None
            Number of timescales to be returned. By default with respect to all available singular values.
        lagtime : int, optional, default=None
            The lagtime with respect to which to compute the timescale. If :code:`None`, this defaults to the
            lagtime under which the covariances were estimated.

        Returns
        -------
        timescales: (n,) np.array
            numpy array with the implied timescales. In principle, one should expect as many timescales as
            input coordinates were available. However, less eigenvalues will be returned if the TICA matrices
            were not full rank or :attr:`dim` contained a floating point percentage, i.e., was interpreted as
            variance cutoff.

        Raises
        ------
        ValueError
            If any of the singular values not real, i.e., has a non-zero imaginary component.
        """
        if not np.all(np.isreal(self.singular_values)):
            raise ValueError("This is only meaningful for real singular values.")
        if lagtime is None:
            lagtime = self._cov.lagtime
        return - lagtime / np.log(np.abs(self.singular_values[:k]))

    @property
    def feature_component_correlation(self):
        r"""Instantaneous correlation matrix between mean-free input features and projection components.

        Denoting the input features as :math:`X_i` and the projection components as :math:`\theta_j`, the
        instantaneous, linear correlation between them can be written as

        .. math::
            \mathbf{Corr}(X_i - \mu_i, \mathbf{\theta}_j) = \frac{1}{\sigma_{X_i - \mu_i}}\sum_l \sigma_{(X_i - \mu_i)(X_l - \mu_l)} \mathbf{U}_{li}

        The matrix :math:`\mathbf{U}` is the matrix containing the eigenvectors of the generalized
        eigenvalue problem as column vectors.

        Returns
        -------
        corr : ndarray(n,m)
            Correlation matrix between input features and projection components. There is a row for each
            feature and a column for each component.
        """
        feature_sigma = np.sqrt(np.diag(self.cov_00))
        return np.dot(self.cov_00, self.singular_vectors_left[:, : self.output_dimension]) / feature_sigma[:, None]


class KoopmanChapmanKolmogorovValidator(LaggedModelValidator):
    r""" Validates a Koopman model

    """

    def __init__(self, test_model: Model, test_estimator: Estimator, test_model_lagtime: int, mlags,
                 observables, statistics, observables_mean_free, statistics_mean_free):
        super().__init__(test_model, test_estimator, test_model_lagtime, mlags)
        self.observables = observables
        self.statistics = statistics
        self.observables_mean_free = observables_mean_free
        self.statistics_mean_free = statistics_mean_free

    @property
    def statistics(self):
        return self._statistics

    @statistics.setter
    def statistics(self, value):
        self._statistics = value
        if self._statistics is not None:
            self.nsets = min(self.observables.shape[1], self._statistics.shape[1])

    def _compute_observables(self, model: CovarianceKoopmanModel, mlag):
        # todo for lag time 0 we return a matrix of nan, until the correct solution is implemented
        if mlag == 0 or model is None:
            if self.statistics is None:
                return np.zeros(self.observables.shape[1]) + np.nan
            else:
                return np.zeros((self.observables.shape[1], self.statistics.shape[1])) + np.nan
        else:
            return model.expectation(statistics=self.statistics, observables=self.observables, lag_multiple=mlag,
                                     statistics_mean_free=self.statistics_mean_free,
                                     observables_mean_free=self.observables_mean_free)

    def _compute_observables_conf(self, model, mlag, conf=0.95):
        raise NotImplementedError('estimation of confidence intervals not yet implemented for Koopman models')



from typing import Union

import numpy as np


def visited_set(dtrajs):
    r"""returns the set of states that have at least one count

    Parameters
    ----------
    dtrajs : array_like or list of array_like
        Discretized trajectory or list of discretized trajectories

    Returns
    -------
    vis : ndarray((n), dtype=int)
        the set of states that have at least one count.
    """
    from deeptime.util.types import ensure_dtraj_list
    dtrajs = ensure_dtraj_list(dtrajs)
    hist = count_states(dtrajs)
    return np.argwhere(hist > 0)[:, 0]


def count_states(dtrajs, ignore_negative: bool = False):
    r"""Computes a histogram over the visited states in one or multiple discretized trajectories.

    Parameters
    ----------
    dtrajs : array_like or list of array_like
        Discretized trajectory or list of discretized trajectories
    ignore_negative : bool, default=False
        Ignore negative elements. By default, a negative element will cause an
        exception

    Returns
    -------
    count : ndarray((n), dtype=int)
        the number of occurrences of each state. n=max+1 where max is the largest state index found.

    """
    from deeptime.util.types import ensure_dtraj_list
    dtrajs = ensure_dtraj_list(dtrajs)

    max_n_states = 0
    histograms = []
    for discrete_trajectory in dtrajs:
        if ignore_negative:
            discrete_trajectory = discrete_trajectory[np.where(discrete_trajectory >= 0)]
        trajectory_histogram = np.bincount(discrete_trajectory)
        max_n_states = max(max_n_states, trajectory_histogram.shape[0])
        histograms.append(trajectory_histogram)
    # allocate space for histogram
    res = np.zeros(max_n_states, dtype=int)
    # aggregate histograms over trajectories
    for trajectory_histogram in histograms:
        res[:trajectory_histogram.shape[0]] += trajectory_histogram
    return res


def compute_effective_stride(dtrajs, lagtime, n_states) -> int:
    r"""
    Computes the effective stride which is an estimate of the striding required to produce uncorrelated samples.
    By default this is the lagtime (lag sampling). A nonreversible MSM is estimated, if its number of states is larger
    than the number of states provided to this method, stride is set to the minimum of lagtime and two times the
    correlation time of the next neglected timescale.

    Parameters
    ----------
    dtrajs : array_like or list of array_like
        Discretized trajectory or list of discretized trajectories
    lagtime : int
        Lagtime
    n_states : int
        Number of resolved states

    Returns
    -------
    stride : int
        Estimated effective stride to produce approximately uncorrelated samples
    """
    from deeptime.util.types import ensure_dtraj_list
    dtrajs = ensure_dtraj_list(dtrajs)
    # by default use lag as stride (=lag sampling), because we currently have no better theory for deciding
    # how many uncorrelated counts we can make
    stride = lagtime
    # get a quick fit from the spectral radius of the non-reversible
    from deeptime.markov import TransitionCountEstimator
    count_model = TransitionCountEstimator(lagtime=lagtime, count_mode="sliding").fit(dtrajs).fetch_model()
    count_model = count_model.submodel_largest()
    from deeptime.markov.msm import MaximumLikelihoodMSM
    msm_non_rev = MaximumLikelihoodMSM(reversible=False, sparse=False).fit(count_model).fetch_model()
    # if we have more than n_states timescales in our MSM, we use the next (neglected) timescale as an
    # fit of the de-correlation time
    if msm_non_rev.n_states > n_states:
        # because we use non-reversible msm, we want to silence the ImaginaryEigenvalueWarning
        import warnings
        with warnings.catch_warnings():
            from deeptime.util.exceptions import ImaginaryEigenValueWarning
            warnings.filterwarnings('ignore', category=ImaginaryEigenValueWarning)
            correlation_time = max(1, msm_non_rev.timescales()[n_states - 1])
        # use the smaller of these two pessimistic estimates
        stride = int(min(lagtime, 2 * correlation_time))

    return stride


def lag_observations(observations, lag, stride=1):
    r""" Create new trajectories that are subsampled at lag but shifted

    Given a trajectory (s0, s1, s2, s3, s4, ...) and lag 3, this function will generate 3 trajectories
    (s0, s3, s6, ...), (s1, s4, s7, ...) and (s2, s5, s8, ...). Use this function in order to parametrize a MLE
    at lag times larger than 1 without discarding data. Do not use this function for Bayesian estimators, where
    data must be given such that subsequent transitions are uncorrelated.

    Parameters
    ----------
    observations : array_like or list of array_like
        observation trajectories
    lag : int
        lag time
    stride : int, default=1
        will return only one trajectory for every stride. Use this for Bayesian analysis.

    """
    # todo cppify
    from deeptime.util.types import ensure_timeseries_data
    observations = ensure_timeseries_data(observations)
    obsnew = []
    for obs in observations:
        for shift in range(0, lag, stride):
            obs_lagged = obs[shift::lag]
            if len(obs_lagged) > 1:
                obsnew.append(obs_lagged)
    return obsnew


def compute_dtrajs_effective(dtrajs, lagtime: int, n_states: int, stride: Union[int, str]):
    r"""
    Takes discrete trajectories as input and strides these with an effective stride. See methods
    `compute_effective_stride` and `lag_observations`.

    Parameters
    ----------
    dtrajs : array_like or list of array_like
        discrete trajectories
    lagtime : int
        lagtime
    n_states : int
        number of resolved states
    stride : int or str
        if 'effective', computes effective stride, otherwise uses int value

    Returns
    -------
    Lagged and stridden observations.
    """
    lagtime = int(lagtime)
    # EVALUATE STRIDE
    if stride == 'effective':
        stride = compute_effective_stride(dtrajs, lagtime, n_states)

    # LAG AND STRIDE DATA
    dtrajs_lagged_strided = lag_observations(dtrajs, lagtime, stride=stride)
    return dtrajs_lagged_strided


def compute_connected_sets(count_matrix, connectivity_threshold: float = 0, directed=True):
    """ Computes the connected sets of a count matrix C.

    C : (N, N) np.ndarray
        count matrix
    mincount_connectivity : float
        Minimum count required to be included in the connected set computation.
    directed : boolean
        True: Seek connected sets in the directed graph. False: Seek connected sets in the undirected graph.
    Returns
    -------
    A list of arrays, each array representing a connected set by enumerating the respective states. The list is in
    descending order by size of connected set.
    """
    import deeptime.markov.tools.estimation as msmest
    import scipy.sparse as scs
    if connectivity_threshold > 0:
        if scs.issparse(count_matrix):
            Cconn = count_matrix.tocsr(copy=True)
            Cconn.data[Cconn.data < connectivity_threshold] = 0
            Cconn.eliminate_zeros()
        else:
            Cconn = count_matrix.copy()
            Cconn[np.where(Cconn < connectivity_threshold)] = 0
    else:
        Cconn = count_matrix
    # treat each connected set separately
    S = msmest.connected_sets(Cconn, directed=directed)
    return S


def number_of_states(dtrajs, only_used=False) -> int:
    r"""Returns the number of states in the given trajectories.

    Parameters
    ----------
    dtrajs : array_like or list of array_like
        Discretized trajectory or list of discretized trajectories
    only_used : bool, default=False
        If False, will return max+1, where max is the largest index used.
        If True, will return the number of states that occur at least once.

    Returns
    -------
    count : int
        The number of states
    """
    from deeptime.util.types import ensure_dtraj_list
    dtrajs = ensure_dtraj_list(dtrajs)
    if only_used:
        # only states with counts > 0 wanted. Make a bincount and count nonzeros
        bc = count_states(dtrajs)
        return np.count_nonzero(bc)
    else:
        # all states wanted, included nonpopulated ones. return max + 1
        return max(np.max(dtraj) for dtraj in dtrajs) + 1


def closed_sets(count_matrix, connectivity_threshold: float = 0):
    r"""
    todo
    Computes the strongly connected closed sets of C
    Parameters
    ----------
    count_matrix
    connectivity_threshold

    Returns
    -------

    """
    n = np.shape(count_matrix)[0]
    S = compute_connected_sets(count_matrix, connectivity_threshold=connectivity_threshold, directed=True)
    closed = []
    for s in S:
        mask = np.zeros(n, dtype=bool)
        mask[s] = True
        if count_matrix[np.ix_(mask, ~mask)].sum() == 0:
            # closed set, take it
            closed.append(s)
    return closed


def is_connected(count_matrix, connectivity_threshold: float = 0, directed=True):
    r"""
    todo
    Parameters
    ----------
    count_matrix
    connectivity_threshold
    directed

    Returns
    -------

    """
    sets = compute_connected_sets(count_matrix, connectivity_threshold=connectivity_threshold, directed=directed)
    return len(sets) == 1



from typing import Optional, Union, Callable, Tuple

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import Dataset

from ...base import Model, Transformer, EstimatorTransformer
from ...base_torch import DLEstimatorMixin
from ...util.torch import map_data


def symeig_reg(mat, epsilon: float = 1e-6, mode='regularize', eigenvectors=True) \
        -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
    r""" Solves a eigenvector/eigenvalue decomposition for a hermetian matrix also if it is rank deficient.

    Parameters
    ----------
    mat : torch.Tensor
        the hermetian matrix
    epsilon : float, default=1e-6
        Cutoff for eigenvalues.
    mode : str, default='regularize'
        Whether to truncate eigenvalues if they are too small or to regularize them by taking the absolute value
        and adding a small positive constant. :code:`trunc` leads to truncation, :code:`regularize` leads to epsilon
        being added to the eigenvalues after taking the absolute value
    eigenvectors : bool, default=True
        Whether to compute eigenvectors.

    Returns
    -------
    (eigval, eigvec) : Tuple[torch.Tensor, Optional[torch.Tensor]]
        Eigenvalues and -vectors.
    """
    assert mode in sym_inverse.valid_modes, f"Invalid mode {mode}, supported are {sym_inverse.valid_modes}"

    if mode == 'regularize':
        identity = torch.eye(mat.shape[0], dtype=mat.dtype, device=mat.device)
        mat = mat + epsilon * identity

    # Calculate eigvalues and potentially eigvectors
    eigval, eigvec = torch.symeig(mat, eigenvectors=True)

    if eigenvectors:
        eigvec = eigvec.transpose(0, 1)

    if mode == 'trunc':
        # Filter out Eigenvalues below threshold and corresponding Eigenvectors
        mask = eigval > epsilon
        eigval = eigval[mask]
        if eigenvectors:
            eigvec = eigvec[mask]
    elif mode == 'regularize':
        # Calculate eigvalues and eigvectors
        eigval = torch.abs(eigval)
    elif mode == 'clamp':
        eigval = torch.clamp_min(eigval, min=epsilon)

    return eigval, eigvec


def sym_inverse(mat, epsilon: float = 1e-6, return_sqrt=False, mode='regularize'):
    """ Utility function that returns the inverse of a matrix, with the
    option to return the square root of the inverse matrix.

    Parameters
    ----------
    mat: numpy array with shape [m,m]
        Matrix to be inverted.
    epsilon : float
        Cutoff for eigenvalues.
    return_sqrt: bool, optional, default = False
        if True, the square root of the inverse matrix is returned instead
    mode: str, default='trunc'
        Whether to truncate eigenvalues if they are too small or to regularize them by taking the absolute value
        and adding a small positive constant. :code:`trunc` leads to truncation, :code:`regularize` leads to epsilon
        being added to the eigenvalues after taking the absolute value

    Returns
    -------
    x_inv: numpy array with shape [m,m]
        inverse of the original matrix
    """
    eigval, eigvec = symeig_reg(mat, epsilon, mode)

    # Build the diagonal matrix with the filtered eigenvalues or square
    # root of the filtered eigenvalues according to the parameter
    if return_sqrt:
        diag = torch.diag(torch.sqrt(1. / eigval))
    else:
        diag = torch.diag(1. / eigval)

    return torch.chain_matmul(eigvec.t(), diag, eigvec)


sym_inverse.valid_modes = ('trunc', 'regularize', 'clamp')


def koopman_matrix(x: torch.Tensor, y: torch.Tensor, epsilon: float = 1e-6, mode: str = 'trunc',
                   c_xx: Optional[Tuple[torch.Tensor, torch.Tensor, torch.Tensor]] = None) -> torch.Tensor:
    r""" Computes the Koopman matrix

    .. math:: K = C_{00}^{-1/2}C_{0t}C_{tt}^{-1/2}

    based on data over which the covariance matrices :math:`C_{\cdot\cdot}` are computed.

    Parameters
    ----------
    x : torch.Tensor
        Instantaneous data.
    y : torch.Tensor
        Time-lagged data.
    epsilon : float, default=1e-6
        Cutoff parameter for small eigenvalues.
    mode : str, default='trunc'
        Regularization mode for Hermetian inverse. See :meth:`sym_inverse`.
    c_xx : tuple of torch.Tensor, optional, default=None
        Tuple containing c00, c0t, ctt if already computed.

    Returns
    -------
    K : torch.Tensor
        The Koopman matrix.
    """
    if c_xx is not None:
        c00, c0t, ctt = c_xx
    else:
        c00, c0t, ctt = covariances(x, y, remove_mean=True)
    c00_sqrt_inv = sym_inverse(c00, return_sqrt=True, epsilon=epsilon, mode=mode)
    ctt_sqrt_inv = sym_inverse(ctt, return_sqrt=True, epsilon=epsilon, mode=mode)
    return torch.chain_matmul(c00_sqrt_inv, c0t, ctt_sqrt_inv).t()


def covariances(x: torch.Tensor, y: torch.Tensor, remove_mean: bool = True):
    """Computes instantaneous and time-lagged covariances matrices.

    Parameters
    ----------
    x : (T, n) torch.Tensor
        Instantaneous data.
    y : (T, n) torch.Tensor
        Time-lagged data.
    remove_mean: bool, default=True
        Whether to remove the mean of x and y.

    Returns
    -------
    cov_00 : (n, n) torch.Tensor
        Auto-covariance matrix of x.
    cov_0t : (n, n) torch.Tensor
        Cross-covariance matrix of x and y.
    cov_tt : (n, n) torch.Tensor
        Auto-covariance matrix of y.

    See Also
    --------
    deeptime.covariance.Covariance : Estimator yielding these kind of covariance matrices based on raw numpy arrays
                                     using an online estimation procedure.
    """

    assert x.shape == y.shape, "x and y must be of same shape"
    batch_size = x.shape[0]

    if remove_mean:
        x = x - x.mean(dim=0, keepdim=True)
        y = y - y.mean(dim=0, keepdim=True)

    # Calculate the cross-covariance
    y_t = y.transpose(0, 1)
    x_t = x.transpose(0, 1)
    cov_01 = 1 / (batch_size - 1) * torch.matmul(x_t, y)
    # Calculate the auto-correlations
    cov_00 = 1 / (batch_size - 1) * torch.matmul(x_t, x)
    cov_11 = 1 / (batch_size - 1) * torch.matmul(y_t, y)

    return cov_00, cov_01, cov_11


valid_score_methods = ('VAMP1', 'VAMP2', 'VAMPE')


def vamp_score(data: torch.Tensor, data_lagged: torch.Tensor, method='VAMP2', epsilon: float = 1e-6, mode='trunc'):
    r"""Computes the VAMP score based on data and corresponding time-shifted data.

    Parameters
    ----------
    data : torch.Tensor
        (N, d)-dimensional torch tensor
    data_lagged : torch.Tensor
        (N, k)-dimensional torch tensor
    method : str, default='VAMP2'
        The scoring method. See :meth:`score <deeptime.decomposition.CovarianceKoopmanModel.score>` for details.
    epsilon : float, default=1e-6
        Cutoff parameter for small eigenvalues, alternatively regularization parameter.
    mode : str, default='trunc'
        Regularization mode for Hermetian inverse. See :meth:`sym_inverse`.

    Returns
    -------
    score : torch.Tensor
        The score. It contains a contribution of :math:`+1` for the constant singular function since the
        internally estimated Koopman operator is defined on a decorrelated basis set.
    """
    assert method in valid_score_methods, f"Invalid method '{method}', supported are {valid_score_methods}"
    assert data.shape == data_lagged.shape, f"Data and data_lagged must be of same shape but were {data.shape} " \
                                            f"and {data_lagged.shape}."
    out = None
    if method == 'VAMP1':
        koopman = koopman_matrix(data, data_lagged, epsilon=epsilon, mode=mode)
        out = torch.norm(koopman, p='nuc')
    elif method == 'VAMP2':
        koopman = koopman_matrix(data, data_lagged, epsilon=epsilon, mode=mode)
        out = torch.pow(torch.norm(koopman, p='fro'), 2)
    elif method == 'VAMPE':
        c00, c0t, ctt = covariances(data, data_lagged, remove_mean=True)
        c00_sqrt_inv = sym_inverse(c00, epsilon=epsilon, return_sqrt=True, mode=mode)
        ctt_sqrt_inv = sym_inverse(ctt, epsilon=epsilon, return_sqrt=True, mode=mode)
        koopman = torch.chain_matmul(c00_sqrt_inv, c0t, ctt_sqrt_inv).t()

        u, s, v = torch.svd(koopman)
        mask = s > epsilon

        u = torch.mm(c00_sqrt_inv, u[:, mask])
        v = torch.mm(ctt_sqrt_inv, v[:, mask])
        s = s[mask]

        u_t = u.t()
        v_t = v.t()
        s = torch.diag(s)

        out = torch.trace(
            2. * torch.chain_matmul(s, u_t, c0t, v)
            - torch.chain_matmul(s, u_t, c00, u, s, v_t, ctt, v)
        )
    assert out is not None
    return 1 + out


def vampnet_loss(data: torch.Tensor, data_lagged: torch.Tensor, method='VAMP2', epsilon: float = 1e-6,
                 mode: str = 'trunc'):
    r"""Loss function that can be used to train VAMPNets. It evaluates as :math:`-\mathrm{score}`. The score
    is implemented in :meth:`score`."""
    return -1. * vamp_score(data, data_lagged, method=method, epsilon=epsilon, mode=mode)


class VAMPNetModel(Transformer, Model):
    r"""
    A VAMPNet model which can be fit to data optimizing for one of the implemented VAMP scores.

    Parameters
    ----------
    lobe : torch.nn.Module
        One of the lobes of the VAMPNet. See also :class:`deeptime.util.torch.MLP`.
    lobe_timelagged : torch.nn.Module, optional, default=None
        The timelagged lobe. Can be left None, in which case the lobes are shared.
    dtype : data type, default=np.float32
        The data type for which operations should be performed. Leads to an appropriate cast within fit and
        transform methods.
    device : device, default=None
        The device for the lobe(s). Can be None which defaults to CPU.

    See Also
    --------
    VAMPNet : The corresponding estimator.
    """

    def __init__(self, lobe: nn.Module, lobe_timelagged: Optional[nn.Module] = None,
                 dtype=np.float32, device=None):
        super().__init__()
        self._lobe = lobe
        self._lobe_timelagged = lobe_timelagged if lobe_timelagged is not None else lobe

        if dtype == np.float32:
            self._lobe = self._lobe.float()
            self._lobe_timelagged = self._lobe_timelagged.float()
        elif dtype == np.float64:
            self._lobe = self._lobe.double()
            self._lobe_timelagged = self._lobe_timelagged.double()
        self._dtype = dtype
        self._device = device

    @property
    def lobe(self) -> nn.Module:
        r""" The instantaneous lobe.

        Returns
        -------
        lobe : nn.Module
        """
        return self._lobe

    @property
    def lobe_timelagged(self) -> nn.Module:
        r""" The timelagged lobe. Might be equal to :attr:`lobe`.

        Returns
        -------
        lobe_timelagged : nn.Module
        """
        return self._lobe_timelagged

    def transform(self, data, instantaneous: bool = True, **kwargs):
        r""" Transforms data through the instantaneous or time-shifted network lobe.

        Parameters
        ----------
        data : numpy array or torch tensor
            The data to transform.
        instantaneous : bool, default=True
            Whether to use the instantaneous lobe or the time-shifted lobe for transformation.
        **kwargs
            Ignored kwargs for api compatibility.

        Returns
        -------
        transform : array_like
            List of numpy array or numpy array containing transformed data.
        """
        if instantaneous:
            self._lobe.eval()
            net = self._lobe
        else:
            self._lobe_timelagged.eval()
            net = self._lobe_timelagged
        out = []
        for data_tensor in map_data(data, device=self._device, dtype=self._dtype):
            out.append(net(data_tensor).cpu().numpy())
        return out if len(out) > 1 else out[0]


class VAMPNet(EstimatorTransformer, DLEstimatorMixin):
    r""" Implementation of VAMPNets. :footcite:`mardt2018vampnets`
    These networks try to find an optimal featurization of data based on a VAMP score :footcite:`wu2020variational`
    by using neural networks as featurizing transforms which are equipped with a loss that is the negative VAMP score.
    This estimator is also a transformer and can be used to transform data into the optimized space.
    From there it can either be used to estimate Markov state models via making assignment probabilities
    crisp (in case of softmax output distributions) or to estimate the Koopman operator
    using the :class:`VAMP <deeptime.decomposition.VAMP>` estimator.

    Parameters
    ----------
    lobe : torch.nn.Module
        A neural network module which maps input data to some (potentially) lower-dimensional space.
    lobe_timelagged : torch.nn.Module, optional, default=None
        Neural network module for timelagged data, in case of None the lobes are shared (structure and weights).
    device : torch device, default=None
        The device on which the torch modules are executed.
    optimizer : str or Callable, default='Adam'
        An optimizer which can either be provided in terms of a class reference (like `torch.optim.Adam`) or
        a string (like `'Adam'`). Defaults to Adam.
    learning_rate : float, default=5e-4
        The learning rate of the optimizer.
    score_method : str, default='VAMP2'
        The scoring method which is used for optimization.
    score_mode : str, default='regularize'
        The mode under which inverses of positive semi-definite matrices are estimated. Per default, the matrices
        are perturbed by a small constant added to the diagonal. This makes sure that eigenvalues are not too
        small. For a complete list of modes, see :meth:`sym_inverse`.
    epsilon : float, default=1e-6
        The strength of the regularization under which matrices are inverted. Meaning depends on the score_mode,
        see :meth:`sym_inverse`.
    dtype : dtype, default=np.float32
        The data type of the modules and incoming data.

    See Also
    --------
    deeptime.decomposition.VAMP

    References
    ----------
    .. footbibliography::
    """
    _MUTABLE_INPUT_DATA = True

    def __init__(self, lobe: nn.Module, lobe_timelagged: Optional[nn.Module] = None,
                 device=None, optimizer: Union[str, Callable] = 'Adam', learning_rate: float = 5e-4,
                 score_method: str = 'VAMP2', score_mode: str = 'regularize', epsilon: float = 1e-6,
                 dtype=np.float32):
        super().__init__()
        self.lobe = lobe
        self.lobe_timelagged = lobe_timelagged
        self.score_method = score_method
        self.score_mode = score_mode
        self._step = 0
        self._epsilon = epsilon
        self.device = device
        self.learning_rate = learning_rate
        self.dtype = dtype
        self.setup_optimizer(optimizer, list(self.lobe.parameters()) + list(self.lobe_timelagged.parameters()))
        self._train_scores = []
        self._validation_scores = []

    @property
    def train_scores(self) -> np.ndarray:
        r""" The collected train scores. First dimension contains the step, second dimension the score. Initially empty.

        :type: (T, 2) ndarray
        """
        return np.array(self._train_scores)

    @property
    def validation_scores(self) -> np.ndarray:
        r""" The collected validation scores. First dimension contains the step, second dimension the score.
        Initially empty.

        :type: (T, 2) ndarray
        """
        return np.array(self._validation_scores)

    @property
    def epsilon(self) -> float:
        r""" Regularization parameter for matrix inverses.

        :getter: Gets the currently set parameter.
        :setter: Sets a new parameter. Must be non-negative.
        :type: float
        """
        return self._epsilon

    @epsilon.setter
    def epsilon(self, value: float):
        assert value >= 0
        self._epsilon = value

    @property
    def score_method(self) -> str:
        r""" Property which steers the scoring behavior of this estimator.

        :getter: Gets the current score.
        :setter: Sets the score to use.
        :type: str
        """
        return self._score_method

    @score_method.setter
    def score_method(self, value: str):
        assert value in valid_score_methods, f"Tried setting an unsupported scoring method '{value}', " \
                                             f"available are {valid_score_methods}."
        self._score_method = value

    @property
    def lobe(self) -> nn.Module:
        r""" The instantaneous lobe of the VAMPNet.

        :getter: Gets the instantaneous lobe.
        :setter: Sets a new lobe.
        :type: torch.nn.Module
        """
        return self._lobe

    @lobe.setter
    def lobe(self, value: nn.Module):
        self._lobe = value
        if self.dtype == np.float32:
            self._lobe = self._lobe.float()
        else:
            self._lobe = self._lobe.double()
        self._lobe = self._lobe.to(device=self.device)

    @property
    def lobe_timelagged(self) -> nn.Module:
        r""" The timelagged lobe of the VAMPNet.

        :getter: Gets the timelagged lobe. Can be the same a the instantaneous lobe.
        :setter: Sets a new lobe. Can be None, in which case the instantaneous lobe is shared.
        :type: torch.nn.Module
        """
        return self._lobe_timelagged

    @lobe_timelagged.setter
    def lobe_timelagged(self, value: Optional[nn.Module]):
        if value is None:
            value = self.lobe
        else:
            if self.dtype == np.float32:
                value = value.float()
            else:
                value = value.double()
        self._lobe_timelagged = value
        self._lobe_timelagged = self._lobe_timelagged.to(device=self.device)

    def partial_fit(self, data, train_score_callback: Callable[[int, torch.Tensor], None] = None):
        r""" Performs a partial fit on data. This does not perform any batching.

        Parameters
        ----------
        data : tuple or list of length 2, containing instantaneous and timelagged data
            The data to train the lobe(s) on.
        train_score_callback : callable, optional, default=None
            An optional callback function which is evaluated after partial fit, containing the current step
            of the training (only meaningful during a :meth:`fit`) and the current score as torch Tensor.

        Returns
        -------
        self : VAMPNet
            Reference to self.
        """

        if self.dtype == np.float32:
            self._lobe = self._lobe.float()
            self._lobe_timelagged = self._lobe_timelagged.float()
        elif self.dtype == np.float64:
            self._lobe = self._lobe.double()
            self._lobe_timelagged = self._lobe_timelagged.double()

        self.lobe.train()
        self.lobe_timelagged.train()

        assert isinstance(data, (list, tuple)) and len(data) == 2, \
            "Data must be a list or tuple of batches belonging to instantaneous " \
            "and respective time-lagged data."

        batch_0, batch_t = data[0], data[1]

        if isinstance(data[0], np.ndarray):
            batch_0 = torch.from_numpy(data[0].astype(self.dtype)).to(device=self.device)
        if isinstance(data[1], np.ndarray):
            batch_t = torch.from_numpy(data[1].astype(self.dtype)).to(device=self.device)

        self.optimizer.zero_grad()
        x_0 = self.lobe(batch_0)
        x_t = self.lobe_timelagged(batch_t)
        loss_value = vampnet_loss(x_0, x_t, method=self.score_method, epsilon=self.epsilon, mode=self.score_mode)
        loss_value.backward()
        self.optimizer.step()

        if train_score_callback is not None:
            lval_detached = loss_value.detach()
            train_score_callback(self._step, -lval_detached)
        self._train_scores.append((self._step, (-loss_value).item()))
        self._step += 1

        return self

    def validate(self, validation_data: Tuple[torch.Tensor]) -> torch.Tensor:
        r""" Evaluates the currently set lobe(s) on validation data and returns the value of the configured score.

        Parameters
        ----------
        validation_data : Tuple of torch Tensor containing instantaneous and timelagged data
            The validation data.

        Returns
        -------
        score : torch.Tensor
            The value of the score.
        """
        self.lobe.eval()
        self.lobe_timelagged.eval()

        with torch.no_grad():
            val = self.lobe(validation_data[0])
            val_t = self.lobe_timelagged(validation_data[1])
            score_value = vamp_score(val, val_t, method=self.score_method, mode=self.score_mode, epsilon=self.epsilon)
            return score_value

    def fit(self, data_loader: torch.utils.data.DataLoader, n_epochs=1, validation_loader=None,
            train_score_callback: Callable[[int, torch.Tensor], None] = None,
            validation_score_callback: Callable[[int, torch.Tensor], None] = None,
            progress=None, **kwargs):
        r""" Fits a VampNet on data.

        Parameters
        ----------
        data_loader : torch.utils.data.DataLoader
            The data to use for training. Should yield a tuple of batches representing
            instantaneous and time-lagged samples.
        n_epochs : int, default=1
            The number of epochs (i.e., passes through the training data) to use for training.
        validation_loader : torch.utils.data.DataLoader, optional, default=None
            Validation data, should also be yielded as a two-element tuple.
        train_score_callback : callable, optional, default=None
            Callback function which is invoked after each batch and gets as arguments the current training step
            as well as the score (as torch Tensor).
        validation_score_callback : callable, optional, default=None
            Callback function for validation data. Is invoked after each epoch if validation data is given
            and the callback function is not None. Same as the train callback, this gets the 'step' as well as
            the score.
        progress : context manager, optional, default=None
            Progress bar (eg tqdm), defaults to None.
        **kwargs
            Optional keyword arguments for scikit-learn compatibility

        Returns
        -------
        self : VAMPNet
            Reference to self.
        """
        from deeptime.util.platform import handle_progress_bar
        progress = handle_progress_bar(progress)
        self._step = 0

        # and train
        for epoch in progress(range(n_epochs), desc="VAMPNet epoch", total=n_epochs, leave=False):
            for batch_0, batch_t in data_loader:
                self.partial_fit((batch_0.to(device=self.device), batch_t.to(device=self.device)),
                                 train_score_callback=train_score_callback)
            if validation_loader is not None:
                with torch.no_grad():
                    scores = []
                    for val_batch in validation_loader:
                        scores.append(
                            self.validate((val_batch[0].to(device=self.device), val_batch[1].to(device=self.device)))
                        )
                    mean_score = torch.mean(torch.stack(scores))
                    self._validation_scores.append((self._step, mean_score.item()))
                    if validation_score_callback is not None:
                        validation_score_callback(self._step, mean_score)
        return self

    def fetch_model(self) -> VAMPNetModel:
        r""" Yields the current model. """
        from copy import deepcopy
        lobe = deepcopy(self.lobe)
        if self.lobe == self.lobe_timelagged:
            lobe_timelagged = lobe
        else:
            lobe_timelagged = deepcopy(self.lobe_timelagged)
        return VAMPNetModel(lobe, lobe_timelagged, dtype=self.dtype, device=self.device)



import numpy as _np
from scipy.sparse import issparse


def is_sorted(x: _np.ndarray, order: str = 'asc'):
    r"""Check if x is sorted.

    Parameters
    ----------
    x : ndarray
        input array
    order : str, default='asc'
        One of asc (for ascending) and desc (for descending)

    Returns
    -------
    sorted : bool
        Whether array is sorted.
    """
    assert order in ('asc', 'desc')
    try:
        if _np.issubdtype(x.dtype, _np.unsignedinteger):
            # x is unsigned int array, risk of int underflow in np.diff
            x = _np.int64(x)
    except AttributeError:
        pass
    if order == 'asc':
        return (_np.diff(x) >= 0).all()
    else:
        return (_np.diff(x) <= 0).all()


def is_diagonal_matrix(matrix: _np.ndarray) -> bool:
    r""" Checks whether a provided matrix is a diagonal matrix, i.e., :math:`A = \mathrm{diag}(a_1,\ldots, a_n)`.

    Parameters
    ----------
    matrix : ndarray
        The matrix for which this check is performed.

    Returns
    -------
    is_diagonal : bool
        True if the matrix is a diagonal matrix, otherwise False.
    """
    return _np.all(matrix == _np.diag(_np.diagonal(matrix)))


def is_square_matrix(arr: _np.ndarray) -> bool:
    r""" Determines whether an array is a square matrix. This means that ndim must be 2 and shape[0] must be equal
    to shape[1].

    Parameters
    ----------
    arr : ndarray or sparse array
        The array to check.

    Returns
    -------
    is_square_matrix : bool
        Whether the array is a square matrix.
    """
    return (issparse(arr) or isinstance(arr, _np.ndarray)) and arr.ndim == 2 and arr.shape[0] == arr.shape[1]


def allclose_sparse(A, B, rtol=1e-5, atol=1e-8):
    """
    Compares two sparse matrices in the same matter like numpy.allclose()
    Parameters
    ----------
    A : scipy.sparse matrix
        first matrix to compare
    B : scipy.sparse matrix
        second matrix to compare
    rtol : float
        relative tolerance
    atol : float
        absolute tolerance

    Returns
    -------
    True, if given matrices are equal in bounds of rtol and atol
    False, otherwise

    Notes
    -----
    If the following equation is element-wise True, then allclose returns
    True.

     absolute(`a` - `b`) <= (`atol` + `rtol` * absolute(`b`))

    The above equation is not symmetric in `a` and `b`, so that
    `allclose(a, b)` might be different from `allclose(b, a)` in
    some rare cases.
    """
    A = A.tocsr()
    B = B.tocsr()

    """Shape"""
    same_shape = (A.shape == B.shape)

    """Data"""
    if same_shape:
        diff = (A - B).data
        same_data = _np.allclose(diff, 0.0, rtol=rtol, atol=atol)
        return same_data
    else:
        return False


def drop_nan_rows(arr: _np.ndarray, *args):
    r"""
    Remove rows in all inputs for which `arr` has `_np.nan` entries.

    Parameters
    ----------
    arr : numpy.ndarray
        Array whose rows are checked for nan entries.
        Any rows containing nans are removed from `arr` and all arguments
        passed via `args`.

    *args : variable length argument list of numpy.ndarray
        Additional arrays from which to remove rows.
        Each argument should have the same number of rows as `arr`.
    """
    nan_inds = _np.isnan(arr).any(axis=1)
    return (arr[~nan_inds], *[arg[~nan_inds] for arg in args])  # parenthesis not redundant in py < 3.8



r"""
.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
"""

import numpy as np
from scipy.sparse import issparse


def mydot(A, B):
    r"""Dot-product that can handle dense and sparse arrays

    Parameters
    ----------
    A : numpy ndarray or scipy sparse matrix
        The first factor
    B : numpy ndarray or scipy sparse matrix
        The second factor

    Returns
    C : numpy ndarray or scipy sparse matrix
        The dot-product of A and B

    """
    if issparse(A) :
        return A.dot(B)
    elif issparse(B):
        return (B.T.dot(A.T)).T
    else:
        return np.dot(A, B)



r"""This module implements the transition matrix covariance function

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>

"""

import numpy as np


def tmatrix_cov(C, row=None):
    r"""Covariance tensor for the non-reversible transition matrix ensemble

    Normally the covariance tensor cov(p_ij, p_kl) would carry four indices
    (i,j,k,l). In the non-reversible case rows are independent so that
    cov(p_ij, p_kl)=0 for i not equal to k. Therefore the function will only
    return cov(p_ij, p_ik).

    Parameters
    ----------
    C : (M, M) ndarray
        Count matrix
    row : int (optional)
        If row is given return covariance matrix for specified row only

    Returns
    -------
    cov : (M, M, M) ndarray
        Covariance tensor

    """

    if row is None:
        alpha = C + 1.0  # Dirichlet parameters
        alpha0 = alpha.sum(axis=1)  # Sum of paramters (per row)

        norm = alpha0 ** 2 * (alpha0 + 1.0)

        """Non-normalized covariance tensor"""
        Z = -alpha[:, :, np.newaxis] * alpha[:, np.newaxis, :]

        """Correct-diagonal"""
        ind = np.diag_indices(C.shape[0])
        Z[:, ind[0], ind[1]] += alpha0[:, np.newaxis] * alpha

        """Covariance matrix"""
        cov = Z / norm[:, np.newaxis, np.newaxis]

        return cov

    else:
        alpha = C[row, :] + 1.0
        return dirichlet_covariance(alpha)


def dirichlet_covariance(alpha):
    r"""Covariance matrix for Dirichlet distribution.

    Parameters
    ----------
    alpha : (M, ) ndarray
        Parameters of Dirichlet distribution

    Returns
    -------
    cov : (M, M) ndarray
        Covariance matrix

    """
    alpha0 = alpha.sum()
    norm = alpha0 ** 2 * (alpha0 + 1.0)

    """Non normalized covariance"""
    Z = -alpha[:, np.newaxis] * alpha[np.newaxis, :]

    """Correct diagonal"""
    ind = np.diag_indices(Z.shape[0])
    Z[ind] += alpha0 * alpha

    """Covariance matrix"""
    cov = Z / norm

    return cov


def error_perturbation_single(C, S, R=None):
    r"""Error-perturbation arising from a given sensitivity

    Parameters
    ----------
    C : (M, M) ndarray
        Count matrix
    S : (M, M) ndarray
        Sensitivity matrix
    R : (M, M) ndarray (optional)
        Sensitivity matrix

    Returns
    -------
    var : float
         Variance (covariance) of observable(s)

    """
    cov = tmatrix_cov(C)  # (M, M, M)
    if R is None:
        R = S
    X = S[:, :, np.newaxis] * cov * R[:, np.newaxis, :]
    return X.sum()


def error_perturbation_cov(C, S):
    r"""Error-perturbation arising from a given sensitivity

    Parameters
    ----------
    C : (M, M) ndarray
        Count matrix
    S : (K, M, M) ndarray
        Sensitivity tensor

    Returns
    -------
    X : (K, K) ndarray
        Covariance matrix for given sensitivity

    """
    K = S.shape[0]
    X = np.zeros((K, K))
    cov = tmatrix_cov(C)
    for i in range(K):
        for j in range(K):
            Q = S[i, :, :]
            R = S[j, :, :]
            X[i, j] = (Q[:, :, np.newaxis] * cov * R[:, np.newaxis, :]).sum()
    return X


def error_perturbation(C, S):
    r"""Error perturbation for given sensitivity matrix.

    Parameters
    ----------
    C : (M, M) ndarray
        Count matrix
    S : (M, M) ndarray or (K, M, M) ndarray
        Sensitivity matrix (for scalar observable) or sensitivity
        tensor for vector observable

    Returns
    -------
    X : float or (K, K) ndarray
        error-perturbation (for scalar observables) or covariance matrix
        (for vector-valued observable)

    """
    if len(S.shape) == 2:  # Scalar observable
        return error_perturbation_single(C, S)
    elif len(S.shape) == 3:  # Vector observable
        return error_perturbation_cov(C, S)
    else:
        raise ValueError("Sensitivity matrix S has to be a 2d or 3d array")



r"""

============
MSM analysis
============

.. currentmodule:: deeptime.markov.tools.analysis

This module (:mod:`deeptime.markov.tools.analysis`) contains functions to analyze a created Markov model, which is
specified with a transition matrix T.

Validation
==========

.. autosummary::
   :toctree: generated/
   :template: class_nomodule.rst

   is_transition_matrix - Positive entries and rows sum to one
   is_rate_matrix - Nonpositive off-diagonal entries and rows sum to zero
   is_connected - Irreducible matrix
   is_reversible - Symmetric with respect to some probability vector pi

Decomposition
=============

Decomposition routines use the scipy LAPACK bindings for dense
numpy-arrays and the ARPACK bindings for scipy sparse matrices.

.. autosummary::
   :toctree: generated/
   :template: class_nomodule.rst

   stationary_distribution - Invariant vector from eigendecomposition
   eigenvalues - Spectrum via eigenvalue decomposition
   eigenvectors - Right or left eigenvectors
   rdl_decomposition - Full decomposition into eigenvalues and eigenvectors
   timescales - Implied timescales from eigenvalues

Expected counts
=================

.. autosummary::
   :toctree: generated/

   expected_counts - Count matrix expected for given initial distribution
   expected_counts_stationary - Count matrix expected for equilibrium distribution

Passage times
=============

.. autosummary::
   :toctree: generated/

   mfpt - Mean first-passage time

Committors and PCCA
===================

.. autosummary::
   :toctree: generated/

   committor - Forward and backward committor
   pcca_memberships - Perron cluster center analysis
   hitting_probability

Fingerprints
============

.. autosummary::
   :toctree: generated/

   fingerprint_correlation
   fingerprint_relaxation
   expectation - Equilibrium expectation value of an observable
   correlation
   relaxation

Sensitivity analysis
====================

.. autosummary::
   :toctree: generated/

   stationary_distribution_sensitivity
   eigenvalue_sensitivity
   timescale_sensitivity
   eigenvector_sensitivity
   mfpt_sensitivity
   committor_sensitivity
   expectation_sensitivity

"""

from ._api import is_transition_matrix, is_rate_matrix, is_connected, is_reversible
from ._api import stationary_distribution, eigenvalues, eigenvectors, rdl_decomposition, timescales, \
    timescales_from_eigenvalues
from ._api import expected_counts, expected_counts_stationary
from ._api import mfpt
from ._api import committor, pcca_memberships, hitting_probability
from ._api import fingerprint_correlation, fingerprint_relaxation, expectation, correlation, relaxation
from ._api import stationary_distribution_sensitivity, eigenvalue_sensitivity, eigenvector_sensitivity, \
    timescale_sensitivity, mfpt_sensitivity, committor_sensitivity, expectation_sensitivity



r"""
.. currentmodule: deeptime.basis

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    Observable
    Concatenation

    Identity
    Monomials
"""

from ._base import Observable, Concatenation
from ._monomials import Identity, Monomials



r"""
.. moduleauthor:: Fabian Paul <fab@zedat.fu-berlin.de>
"""

import logging
import warnings
import numpy as np
import scipy as sp
import scipy.sparse

from scipy.optimize import fmin_l_bfgs_b
from scipy.special import exprel

from deeptime.util.exceptions import NotConvergedWarning, NotConvergedError
from ...kahandot import kdot, ksum
from ...estimation import transition_matrix
from ...analysis import stationary_distribution


__all__ = [
    'PseudoGeneratorEstimator',
    'TruncatedLogarithmEstimator',
    'CrommelinVandenEijndenEstimator',
    'KalbfleischLawlessEstimator',
    'NotConnectedWarning',
    'estimate_rate_matrix'
]


class NotConnectedWarning(UserWarning):
    pass


def getV(lam, tau):
    n = len(lam)
    V = np.zeros((n, n))
    ones = np.ones((n, n))
    delta = tau * (lam - lam[:, np.newaxis])  # delta_ij = lam_j-lam_i
    positive = delta >= 0
    negative = delta < 0
    a1 = tau * np.exp(tau * lam)[:, np.newaxis] * ones  # tau*e^l_i
    b1 = exprel(delta)
    V[negative] = a1[negative] * b1[negative]
    a2 = tau * np.exp(tau * lam)[np.newaxis, :] * ones  # tau*e^l_j *
    b2 = exprel(-delta)
    V[positive] = a2[positive] * b2[positive]
    return V


def eigen_decomposition(M, pi):
    # eigen decomposition for reversible transition matrix
    sqrt_pi = np.sqrt(pi)
    Msym = sqrt_pi[:, np.newaxis] * M / sqrt_pi  # Msym_ij = M_ij sqrt(pi_i/pi_j)
    lam, B = np.linalg.eigh(Msym)
    A = B / sqrt_pi[:, np.newaxis]  # A_ij = B_ij / sqrt(pi_i)
    Ainv = B.T * sqrt_pi  # Ainv_ij = B_ji * sqrt(pi_j)
    A = np.ascontiguousarray(A)
    Ainv = np.ascontiguousarray(Ainv)
    return lam, A, Ainv


def vdot(a, b):
    # Kahan summation scalar product
    n = a.shape[0]
    return kdot(a.reshape((1, n)), b.reshape((n, 1)))[0, 0]


def sum1(M):
    # 1-D Kahan summation along axis 1
    n, m = M.shape
    x = np.zeros(n)
    for i in range(n):
        x[i] = ksum(np.reshape(M[i, :], (1, m)))
    return x


def raise_or_warn(msg, on_error, warning=UserWarning, exception=RuntimeError):
    if on_error == 'raise':
        raise exception(msg)
    elif on_error == 'warn':
        warnings.warn(msg, warning)
    else:
        raise ValueError('Unsupported value of on_error (%s). Should be "raise" or "warn".'%on_error)


class _RateMatrixEstimator:
    # base class: includes parametrization of K matrix and the basic class interface
    def __init__(self, C, dt=1.0, sparsity=None, t_agg=None, pi=None, tol=1.0E7, maxiter=100000, on_error='raise'):
        if not np.all(C >= 0):
            raise ValueError('Elements of C matrix should be non-negative.')
        if not C.shape[0] == C.shape[1]:
            raise ValueError('C matrix should be square.')
        self.zero_C = np.where(C == 0)
        self.nonzero_C = np.where(C != 0)
        if not dt > 0.0:
            raise ValueError('dt should be positive.')
        if sparsity is not None:
            if not np.all(sparsity >= 0):
                raise ValueError('Elements of sparsity matrix should be non-negative.')
            if not sparsity.shape[0] == sparsity.shape[1] == C.shape[0]:
                raise ValueError('sparsity matrix should be square and of the same dimension as the C matrix.')
        if pi is not None:
            if not np.all(pi > 0.0):
                raise ValueError('pi should be positive.')
            if not pi.shape[0] == C.shape[0]:
                raise ValueError('pi and the C matrix should have compatible dimensions.')
            if not np.allclose(np.sum(pi), 1.0):
                raise ValueError('pi should sum to one.')
        if t_agg is not None:
            if not t_agg > 0:
                raise ValueError('t_agg should be positive.')
            self.t_agg = t_agg
        else:
            self.t_agg = dt*C.sum()

        self.N = C.shape[0]
        self.C = C
        self.pi = pi
        self.dt = dt
        self.sparsity = sparsity
        self.tol = tol
        self.verbose = False
        self.maxiter = maxiter
        self.on_error = on_error


class _ReversibleRateMatrixEstimator(_RateMatrixEstimator):
    # this estimator requires a known stationary vector
    def __init__(self, C, pi, dt=1.0, sparsity=None, t_agg=None, tol=1.0E7, maxiter=100000, on_error='raise'):
        super(_ReversibleRateMatrixEstimator, self).__init__(C, dt=dt, sparsity=sparsity, t_agg=t_agg, pi=pi, tol=tol, maxiter=maxiter, on_error=on_error)

        if self.sparsity is None:
            self.I, self.J = np.triu_indices(self.N, k=1)
            # lower bounds
            self.lower_bounds = np.zeros(len(self.I))
            self.bounds = [(0, None)] * len(self.I)
        else:
            A = self.sparsity + self.sparsity.T
            A[np.tril_indices(self.N, k=0)] = 0
            Asparse = sp.sparse.coo_matrix(A)
            self.I = Asparse.row
            self.J = Asparse.col
            # set lower bounds
            self.bounds = [None] * len(self.I)
            self.lower_bounds = np.zeros(len(self.I))
            for i, j, n in zip(self.I, self.J, range(len(self.I))):
                self.lower_bounds[n] = 1.0 / (self.t_agg * (1.0 / self.pi[i] + 1.0 / self.pi[j]))
                assert self.lower_bounds[n] > 0.0 # self-consitency test
                self.bounds[n] = (self.lower_bounds[n], None)

        # for matrix derivatives
        self.D = [None] * (len(self.I))
        for i, j, n in zip(self.I, self.J, range(len(self.I))):
            Dn = sp.sparse.lil_matrix((self.N, self.N))
            Dn[i, j] = 1.0 / self.pi[i]
            Dn[j, i] = 1.0 / self.pi[j]
            Dn[i, i] = -1.0 / self.pi[i]
            Dn[j, j] = -1.0 / self.pi[j]
            self.D[n] = sp.sparse.coo_matrix(Dn)

    def selftest(self):
        self.count = -1
        x0 = self.initial

        direction = np.random.rand(len(self.initial)) * 1.0E-12
        f1, grad1 = self.function_and_gradient(x0)
        f2 = self.function(x0 + direction)
        df = np.dot(grad1, direction)
        # in general we would have to use |a-b|/max(|a|,|b|) but since mostly |df|>|f1-f2| we can use error/|df|
        err = np.abs((f2 - f1 - df) / (df))
        logging.info('Self-test for rate matrix yields a finite difference of '
                     '%f and a directional derivative of %f. This corresponds '
                     'to a relative error of %f.' % (f2-f1, df, err))

    def run(self):
        """Run the minimization.

        Returns
        -------
        K : (N,N) ndarray
            the optimal rate matrix
        """
        if self.verbose:
            self.selftest()
        self.count = 0
        if self.verbose:
            logging.info('initial value of the objective function is %f'
                         % self.function(self.initial))
        theta0 = self.initial
        theta, f, d = fmin_l_bfgs_b(self.function_and_gradient, theta0, fprime=None, args=(),
                                    approx_grad=False, bounds=self.bounds, factr=self.tol,
                                    pgtol=1.0E-11, disp=0, maxiter=self.maxiter, maxfun=self.maxiter, maxls=100)
        if self.verbose:
            logging.info('l_bfgs_b says: '+str(d))
            logging.info('objective function value reached: %f' % f)
        if d['warnflag'] != 0:
            raise_or_warn(str(d), on_error=self.on_error, warning=NotConvergedWarning, exception=NotConvergedError)

        K = np.zeros((self.N, self.N))
        K[self.I, self.J] = theta / self.pi[self.I]
        K[self.J, self.I] = theta / self.pi[self.J]
        np.fill_diagonal(K, -np.sum(K, axis=1))
        self.K = K
        return K


class PseudoGeneratorEstimator(_RateMatrixEstimator):
    def __init__(self, C, dt, sparsity=None, t_agg=None, pi=None, tol=1.0E7, maxiter=100000, on_error='raise'):
        super(PseudoGeneratorEstimator, self).__init__(C, dt, sparsity=sparsity, t_agg=t_agg, pi=pi, tol=tol, maxiter=maxiter, on_error=on_error)

    def run(self, maxiter=100000, on_error='raise'):
        if self.pi is None:
            self.T = transition_matrix(self.C, reversible=True)
            self.pi = stationary_distribution(self.T)
        else:
            self.T = transition_matrix(self.C, reversible=True, mu=self.pi)

        self.K = (self.T - np.eye(self.N)) / self.dt
        return self.K


class TruncatedLogarithmEstimator(_RateMatrixEstimator):
    def __init__(self, C, dt, sparsity=None, t_agg=None, pi=None, tol=1.0E7, maxiter=100000, on_error='raise'):
        super(TruncatedLogarithmEstimator, self).__init__(C, dt, sparsity=sparsity, t_agg=t_agg, pi=pi, tol=tol, maxiter=maxiter, on_error=on_error)

    def run(self, maxiter=100000, on_error='raise'):
        if self.pi is None:
            self.T = transition_matrix(self.C, reversible=True)
            self.pi = stationary_distribution(self.T)
        else:
            self.T = transition_matrix(self.C, reversible=True, mu=self.pi)

        self.K = np.maximum(np.array(sp.linalg.logm(np.dot(self.T, self.T))/(2.0*self.dt)), 0)
        np.fill_diagonal(self.K, 0)
        np.fill_diagonal(self.K, -sum1(self.K))

        return self.K


class CrommelinVandenEijndenEstimator(_ReversibleRateMatrixEstimator):
    r"""Estimate rate matrix from count matrix using Crommelin-Vanden-Eijnden algorithm.

    The algorithm consists of minimizing the following objective function:

    .. math:: f(K)=\sum_{ij}\left(\sum_{kl}U_{ik}^{-1}K_{kl}U_{lj}-L_{ij}\right)^2\left|\Lambda_{i}\Lambda_{j}\right|

    where :math:`\Lambda_i` are the eigenvalues of :math:`T` and :math:`U` is the matrix
    of its (right) eigenvectors; :math:`L_{ij} = \delta_{ij} \frac{1}{\tau} \log \left| \Lambda_i \right|`.

    :math:`T` is computed from a count matrix using the reversible
    maximum likelihood estimator.

    If sparsity is not None, the maximization is carried out under
    the constraint :math:`K_{ij}=0` and :math:`K_{ji}=0` for all i,j
    where sparsity[i,j]+sparsity[j,i]=0.

    Parameters
    ----------
    C : (N,N) ndarray
        count matrix at a lag time dt
    K0 : (N,N) ndarray
        initial guess for the rate matrix
    pi : (N) ndarray
        the stationary distribution of the desired rate matrix K
    dt : float, optional, default=1.0
        lag time that was used to estimate C
    sparsity : (N,N) ndarray or None, optional, default=None
        If sparsity is None, a fully occupied rate matrix will be estimated.
        Alternatively, with the methods 'CVE' and 'KL' a ndarray of the
        same shape as C can be given. If sparsity[i,j]=0 and sparsity[j,i]=0
        the rate matrix elements :math:`K_{ij}` and :math:`K_{ji}` will be
        constrained to zero.
    t_agg : float, optional
        the aggregated simulation time
        By default this is the total number of transition counts times
        the lag time (no sliding window counting). This value is used
        to compute the lower bound on the transition rate (that are not zero).
        If sparsity is None, this value is ignored.
    tol : float, optional, default = 1.0E7
        Tolerance of the quasi-Newton algorithm that is used to minimize
        the objective function. This is passed as the `factr` parameter to
        `scipy.optimize.fmin_l_bfgs_b`.
        Typical values for factr are: 1e12 for low accuracy; 1e7
        for moderate accuracy; 10.0 for extremely high accuracy.
    maxiter : int, optional, default = 100000
        Minimization of the objective function will do at most this number
        of steps.
    on_error : string, optional, default = 'raise'
        What to do then an error happend. When 'raise' is given, raise
        an exception. When 'warn' is given, produce a (Python) warning.

    Note
    ----
    To compute the rate matrix, call the `run` method of the estimator object.
    """

    def __init__(self, T, K0, pi, dt=1.0, sparsity=None, t_agg=None, tol=1.0E7, maxiter=100000, on_error='raise'):
        from deeptime.markov.tools.analysis import is_transition_matrix

        super(CrommelinVandenEijndenEstimator, self).__init__(T, pi, dt=dt, sparsity=sparsity, t_agg=t_agg, tol=tol, maxiter=maxiter, on_error=on_error)

        if not K0.shape[0] == K0.shape[1] == self.N:
            raise ValueError('Shapes of K0 matrix (initial guess) and count matrix do not match.')
        if not is_transition_matrix(T):
            raise_or_warn('T is not a valid transition matrix.', self.on_error)

        evals, self.U, self.Uinv = eigen_decomposition(T, self.pi)
        if not np.all(np.abs(evals) > 0.0):  # don't allow eigenvalue==exactly zero
            raise ValueError('T has eigenvalues that are exactly zero, can\'t proceed with rate matrix estimation. '
                             'If the CVE method is only used to intitialize the KL method, you might try to call the KL '
                             'method with an initial guess of the rate matrix (K0) instead of intializing with CVE.')
        assert np.allclose(self.Uinv.dot(T).dot(self.U), np.diag(evals))  # self-consistency test

        self.c = np.abs(evals)
        self.L = np.diag(np.log(np.abs(evals)) / self.dt)

        theta = self.pi[self.I] * K0[self.I, self.J]
        self.initial = np.maximum(theta, self.lower_bounds)

    def function(self, x):
        if not np.all(x>=self.lower_bounds):
            raise_or_warn('Optimizer violated the lower bounds for rate matrix elements.', self.on_error)

        # compute K
        K = np.zeros((self.N, self.N))
        K[self.I, self.J] = x / self.pi[self.I]
        K[self.J, self.I] = x / self.pi[self.J]
        np.fill_diagonal(K, -np.sum(K, axis=1))
        # compute function
        delta = self.Uinv.dot(K).dot(self.U) - self.L
        f = self.c.dot((delta * delta).dot(self.c))
        return f

    def function_and_gradient(self, x):
        if not np.all(x>=self.lower_bounds):
            raise_or_warn('Optimizer violated the lower bounds for rate matrix elements.', self.on_error)

        # compute K
        K = np.zeros((self.N, self.N))
        K[self.I, self.J] = x / self.pi[self.I]
        K[self.J, self.I] = x / self.pi[self.J]
        np.fill_diagonal(K, -np.sum(K, axis=1))
        # compute function
        delta = self.Uinv.dot(K).dot(self.U) - self.L
        f = self.c.dot((delta * delta).dot(self.c))

        if self.verbose:
            logging.info('iteration=%d, norm^2=%f' % (self.count, f))

        self.count += 1

        # compute gradient
        X = self.c[:, np.newaxis] * delta * self.c * 2.0
        Y = self.U.dot(X.T).dot(self.Uinv).T
        grad = np.zeros(len(x))
        for i in range(len(x)):
            Di = self.D[i]
            grad[i] = Y[Di.row, Di.col].dot(Di.data)  # sparse version; scipy's sparse multiply oddly returns dense
        return (f, grad)


class KalbfleischLawlessEstimator(_ReversibleRateMatrixEstimator):
    r"""Estimate rate matrix from count matrix using Kalbfleisch-Lawless algorithm.

    The algorithm consists of maximizing the following log-likelihood:

    .. math:: f(K)=\log L=\sum_{ij}C_{ij}\log(e^{K\tau})_{ij}

    where :math:`C_{ij}` are the transition counts at a lag-time :math:`\tau`.
    Here :math:`e` is the matrix exponential and the logarithm is taken element-wise.

    If sparsity is not None, the maximization is carried out under
    the constraint :math:`K_{ij}=0` and :math:`K_{ji}=0` for all i,j
    where sparsity[i,j]+sparsity[j,i]=0.

    Parameters
    ----------
    C : (N,N) ndarray
        count matrix at a lag time dt
    K0 : (N,N) ndarray
        initial guess for the rate matrix
    pi : (N) ndarray
        the stationary distribution of the desired rate matrix K
    dt : float, optional, default=1.0
        lag time that was used to estimate C
    sparsity : (N,N) ndarray or None, optional, default=None
        If sparsity is None, a fully occupied rate matrix will be estimated.
        Alternatively, with the methods 'CVE' and 'KL' a ndarray of the
        same shape as C can be given. If sparsity[i,j]=0 and sparsity[j,i]=0
        the rate matrix elements :math:`K_{ij}` and :math:`K_{ji}` will be
        constrained to zero.
    t_agg : float, optional
        the aggregated simulation time
        By default this is the total number of transition counts times
        the lag time (no sliding window counting). This value is used
        to compute the lower bound on the transition rate (that are not zero).
        If sparsity is None, this value is ignored.
    tol : float, optional, default = 1.0E7
        Tolerance of the quasi-Newton algorithm that is used to minimize
        the objective function. This is passed as the `factr` parameter to
        `scipy.optimize.fmin_l_bfgs_b`.
        Typical values for factr are: 1e12 for low accuracy; 1e7
        for moderate accuracy; 10.0 for extremely high accuracy.
    maxiter : int, optional, default = 100000
        Minimization of the objective function will do at most this number
        of steps.
    on_error : string, optional, default = 'raise'
        What to do then an error happend. When 'raise' is given, raise
        an exception. When 'warn' is given, produce a (Python) warning.

    Note
    ----
    To compute the rate matrix, call the `run` method of the estimator object.
    """

    def __init__(self, C, K0, pi, dt=1.0, sparsity=None, t_agg=None, tol=1.0E7, maxiter=100000, on_error='raise'):
        super(KalbfleischLawlessEstimator, self).__init__(C, pi, dt=dt, sparsity=sparsity, t_agg=t_agg, tol=tol, maxiter=maxiter, on_error=on_error)

        if not K0.shape[0] == K0.shape[1] == self.N:
            raise ValueError('Shapes of K0 matrix (initial guess) and count matrix do not match.')

        # specific variables for KL estimator
        self.sqrt_pi = np.sqrt(pi)

        self.initial = self.pi[self.I] * K0[self.I, self.J]
        self.initial = np.maximum(self.initial, self.lower_bounds)

    def function(self, x):
        self.count += 1
        if not np.all(x>=self.lower_bounds):
            raise_or_warn('Optimizer violated the lower bounds for rate matrix elements.', self.on_error)

        # compute function
        K = np.zeros((self.N, self.N))
        K[self.I, self.J] = x / self.pi[self.I]
        K[self.J, self.I] = x / self.pi[self.J]
        np.fill_diagonal(K, -sum1(K))
        T = sp.linalg.expm(self.dt * K)  # use eigendecomposition instead?
        T[self.zero] = 1.0  # set unused elements to dummy to avoid division by 0
        # check T!=0 for C!=0
        nonzero_C = np.where(self.C != 0)
        if np.any(np.abs(T[nonzero_C]) <= 1E-20):
            warnings.warn('Warning: during iteration T_ij became very small while C(tau)_ij > 0.', NotConnectedWarning)
        f = ksum(self.C * np.log(T))

        if self.verbose:
            logging.info('iteration=%d, log L=%f' % (self.count, f))
        return -f

    def function_and_gradient(self, x):
        if not np.all(x>=self.lower_bounds):
            raise_or_warn('Optimizer violated the lower bounds for rate matrix elements.', self.on_error)

        # compute function
        K = np.zeros((self.N, self.N))
        K[self.I, self.J] = x / self.pi[self.I]
        K[self.J, self.I] = x / self.pi[self.J]
        np.fill_diagonal(K, -sum1(K))

        # compute eigendecomposition
        lam, A, Ainv = eigen_decomposition(K, self.pi)

        # T = kdot(kdot(A,np.diag(np.exp(self.tau*lam))),Ainv)
        T = sp.linalg.expm(self.dt * K)
        T[self.zero_C] = 1.0  # set unused elements to dummy to avoid division by 0
        # check T!=0 for C!=0
        if np.any(np.abs(T[self.nonzero_C]) <= 1E-20):
            warnings.warn('Warning: during iteration T_ij became very small while C(tau)_ij > 0. Regularizing T.', NotConnectedWarning)
            for i,j in zip(*self.nonzero_C):
                if T[i,j] <= 1E-20:
                    T[i,j] = 1E-20

        f = ksum(self.C * np.log(T))

        if self.verbose:
            logging.info('iteration=%d, log L=%f' % (self.count, f))
        self.count += 1

        V = getV(lam, self.dt)

        # M = Ainv.dot(Ctau.T/T.T).dot(A)*V.T
        M = kdot(kdot(Ainv, np.ascontiguousarray(self.C.T / T.T)), A) * np.ascontiguousarray(V.T)
        # H = A.dot(M).dot(Ainv)
        H = kdot(kdot(A, M), Ainv)

        grad = np.zeros(len(x))
        for i in range(len(x)):
            Di = self.D[i]
            grad[i] = vdot(H[Di.col, Di.row], Di.data)  # this is (H.T*Di).sum()

        return (-f, -grad)


def estimate_rate_matrix(C, dt=1.0, method='KL', sparsity=None,
                         t_agg=None, pi=None, tol=1.0E7, K0=None,
                         maxiter=100000, on_error='raise'):
    r"""Estimate a reversible rate matrix from a count matrix.

    Parameters
    ----------
    C : (N,N) ndarray
        count matrix at a lag time dt
    dt : float, optional, default=1.0
        lag time that was used to estimate C
    method : {'KL', 'CVE', 'pseudo', 'truncated_log'}
        Method to use for estimation of the rate matrix.

        * 'pseudo' selects the pseudo-generator. A reversible transition
          matrix T is estimated and (T-Id)/dt is retruned as the rate matrix.

        * 'truncated_log' selects the truncated logarithm [3]_. A
          reversible transition matrix T is estimated and max(logm(T*T)/(2dt),0)
          is returned as the rate matrix. logm is the matrix logarithm and
          the maximum is taken element-wise.

        * 'CVE' selects the algorithm of Crommelin and Vanden-Eijnden [1]_.
          It consists of minimizing the following objective function:

          .. math:: f(K)=\sum_{ij}\left(\sum_{kl} U_{ik}^{-1}K_{kl}U_{lj}-L_{ij}\right)^2 \left|\Lambda_{i}\Lambda_{j}\right|

          where :math:`\Lambda_i` are the eigenvalues of :math:`T` and :math:`U`
          is the matrix of its (right) eigenvectors; :math:`L_{ij}=\delta_{ij}\frac{1}{\tau}\log\left|\Lambda_i\right|`.
          :math:`T` is computed from C using the reversible maximum likelihood
          estimator.

        * 'KL' selects the algorihtm of Kalbfleisch and Lawless [2]_.
          It consists of maximizing the following log-likelihood:

          .. math:: f(K)=\log L=\sum_{ij}C_{ij}\log(e^{K\Delta t})_{ij}

          where :math:`C_{ij}` are the transition counts at a lag-time :math:`\Delta t`.
          Here :math:`e` is the matrix exponential and the logarithm is taken
          element-wise.

    sparsity : (N,N) ndarray or None, optional, default=None
        If sparsity is None, a fully occupied rate matrix will be estimated.
        Alternatively, with the methods 'CVE' and 'KL' a ndarray of the
        same shape as C can be supplied. If sparsity[i,j]=0 and sparsity[j,i]=0
        the rate matrix elements :math:`K_{ij}` and :math:`K_{ji}` will be
        constrained to zero.
    t_agg : float, optional
        the aggregated simulation time;
        by default this is the total number of transition counts times
        the lag time (no sliding window counting). This value is used
        to compute the lower bound on the transition rate (that are not zero).
        If sparsity is None, this value is ignored.
    pi : (N) ndarray, optional
        the stationary vector of the desired rate matrix K.
        If no pi is given, the function takes the stationary vector
        of the MLE reversible T matrix that is computed from C.
    tol : float, optional, default = 1.0E7
        Tolerance of the quasi-Newton algorithm that is used to minimize
        the objective function. This is passed as the `factr` parameter to
        `scipy.optimize.fmin_l_bfgs_b`.
        Typical values for factr are: 1e12 for low accuracy; 1e7
        for moderate accuracy; 10.0 for extremely high accuracy.
    maxiter : int, optional, default = 100000
        Minimization of the objective function will do at most this number
        of steps.
    on_error : string, optional, default = 'raise'
        What to do then an error happend. When 'raise' is given, raise
        an exception. When 'warn' is given, produce a (Python) warning.

    Returns
    -------
    K : (N,N) ndarray
        the optimal rate matrix

    Notes
    -----
    In this implementation the algorithm of Crommelin and Vanden-Eijnden
    (CVE) is initialized with the pseudo-generator estimate. The
    algorithm of Kalbfleisch and Lawless (KL) is initialized using the
    CVE result.

    Example
    -------
    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import rate_matrix
    >>> C = np.array([[100,1],[50,50]])
    >>> R = rate_matrix(C)
    >>> print(np.array_str(R, precision=5))
    [[-0.01385  0.01385]
     [ 0.6993  -0.6993 ]]

    References
    ----------
    .. [1] D. Crommelin and E. Vanden-Eijnden. Data-based inference of
        generators for markov jump processes using convex optimization.
        Multiscale. Model. Sim., 7(4):1751-1778, 2009.
    .. [2] J. D. Kalbfleisch and J. F. Lawless. The analysis of panel
        data under a markov assumption. J. Am. Stat. Assoc.,
        80(392):863-871, 1985.
    .. [3] E. B. Davies. Embeddable Markov Matrices. Electron. J. Probab.
        15:1474, 2010.
    """
    if method not in ['pseudo', 'truncated_log', 'CVE', 'KL']:
        raise ValueError("method must be one of 'KL', 'CVE', 'pseudo' or 'truncated_log'")

    # special case: truncated matrix logarithm
    if method == 'truncated_log':
        e_tlog = TruncatedLogarithmEstimator(C, dt=dt, sparsity=sparsity, t_agg=t_agg, pi=pi, tol=tol, maxiter=maxiter, on_error=on_error)
        e_tlog.run()
        return e_tlog.K

    # remaining algorithms are based on each other in the order pseudo->CVE->KL
    if method == 'pseudo' or method == 'CVE' or K0 is None or pi is None:
        e_pseudo = PseudoGeneratorEstimator(C, dt=dt, sparsity=sparsity, t_agg=t_agg, pi=pi, tol=tol, maxiter=maxiter, on_error=on_error)
        e_pseudo.run()
        if pi is None:
            pi = e_pseudo.pi
        if method == 'pseudo':
            return e_pseudo.K

    if method == 'CVE' or K0 is None:
        if K0 is not None:
            K_init = K0
        else:
            K_init = e_pseudo.K
        e_CVE = CrommelinVandenEijndenEstimator(e_pseudo.T, K_init, pi, dt=dt, sparsity=sparsity, t_agg=t_agg, tol=tol, maxiter=maxiter, on_error=on_error)
        e_CVE.run()
        if method == 'CVE':
            return e_CVE.K

    if K0 is not None:
        K_init = K0
    else:
        K_init = e_CVE.K
    e_KL = KalbfleischLawlessEstimator(C, K_init, pi, dt=dt, sparsity=sparsity, t_agg=t_agg, tol=tol, maxiter=maxiter, on_error=on_error)
    e_KL.run()
    return e_KL.K



import functools
from weakref import WeakKeyDictionary

from deeptime.util.platform import module_available


class cached_property(property):
    r"""
    Property that gets cached, obeys property api and can also be invalidated and overridden. Inspired from
    https://github.com/pydanny/cached-property/ and  https://stackoverflow.com/a/17330273.
    """
    _default_cache_entry = object()

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        super(cached_property, self).__init__(fget, fset, fdel, doc)
        self.cache = WeakKeyDictionary()

    def __get__(self, instance, owner):
        if instance is None:
            return self
        value = self.cache.get(instance, self._default_cache_entry)
        if value is self._default_cache_entry:
            value = self.fget(instance)
            self.cache[instance] = value
        return value

    def __set__(self, instance, value):
        self.cache[instance] = value

    def __delete__(self, instance):
        del self.cache[instance]

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__)

    def invalidate(self):
        self.cache.clear()


def plotting_function(fn):  # pragma: no cover
    r""" Decorator marking a function that is a plotting utility. This will exclude it from coverage and test
    whether dependencies are installed. """

    @functools.wraps(fn)
    def wrapper(*args, **kw):
        if not module_available("matplotlib") or not module_available("networkx"):
            raise RuntimeError("Plotting functions require matplotlib and networkx to be installed.")
        return fn(*args, **kw)

    return wrapper



r"""Transition matrix sampling for non-reversible stochastic matrices.

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>

"""

import numpy as np


class SamplerNonRev:
    def __init__(self, Z, seed: int = -1):
        """Posterior counts"""
        self.Z = Z
        """Alpha parameters for dirichlet sampling"""
        self.alpha = Z + 1.0
        """Initial state from single sample"""
        self.P = np.zeros_like(Z)
        self.rnd = np.random.RandomState(seed if isinstance(seed, int) and seed >= 0 else None)
        self.update()

    def update(self, N=1):
        N = self.alpha.shape[0]
        for i in range(N):
            # only pass positive alphas to dirichlet sampling.
            positive = self.alpha[i, :] > 0
            self.P[i, positive] = np.random.dirichlet(self.alpha[i, positive])

    def sample(self, N=1, return_statdist=False):
        from ....analysis import stationary_distribution

        self.update(N=N)
        if return_statdist:
            pi = stationary_distribution(self.P)
            return self.P, pi
        else:
            return self.P
